# DATA at the end of the file
# NOTE: to convert to aac: afconvert blah.wav -f adts

import os,sys,traceback,math,pygame,shutil
import simplejson

################################################################################
# HELP
################################################################################

"""

ppp - indicates the pixels per point of this item (1,2,3,...)
    default = 1

pad - set the amount of padding around subtextures in an atlas (1,2,3,...)
    default = 1

scale - scale down the texture image by a factor (1,2,4,8...)
    default = 1
    

"""

################################################################################
# MAIN
################################################################################

def main(DATA):
    pygame_init()
    iname = save_as('data.json')
    info = load_info(iname)
    try:
        for name,e in DATA.items():
            target = e.get_target()
            deps = e.get_deps()
            klass = e.__class__.__name__
            if not is_changed(target,deps): continue
            print name,klass,target,deps
            e.name = name
            e.build()
            add_info(info,name,e.get_info())
            for k,v in e.get_children().items():
                add_info(info,k,v)
    except:
        traceback.print_exc()
    save_info(info,iname)

def is_changed(target,_deps):
    deps = _deps[:]
    deps.insert(0,sys.argv[0])
    deps.insert(0,__file__)
    if not os.path.exists(target): return True
    ts = os.stat(target).st_mtime
    for fname in deps:
        if os.stat(fname).st_mtime > ts:
            return True
    return False
    
def add_info(info,k,v):
    info[k]=v
    info[k]['name'] = k

def load_info(fname):
    try:
        f = open(fname,'rb')
    except:
        return {}
    r = simplejson.load(f)
    f.close()
    return r
        

def save_info(r,fname):
    f = open(fname,'wb')
    simplejson.dump(r,f) #,sort_keys=True,indent=4)
    f.close()

################################################################################
# PARSE ARGUMENTS
################################################################################

MODE = 'default'
DESTDIR = './out'
#TMPDIR = './tmp'

def parse_args():
    global MODE,DESTDIR
    #print sys.argv
    if len(sys.argv)<2:
        print 'python %s DESTDIR [MODE]'%(sys.argv[0])
        exit(-1)
    DESTDIR = sys.argv[1]
    print 'DESTDIR = %s'%DESTDIR
    if len(sys.argv) >= 3:
        MODE = sys.argv[2]
        print 'MODE = %s'%(MODE)

################################################################################
# PYGAME STUFF
################################################################################

SW,SH = 512,512
def pygame_init():
    pygame.display.set_mode((SW,SH),0,32)
    pygame.font.init()
def pygame_show(img):
    screen = pygame.display.get_surface()
    screen.fill((255,0,255))
    screen.blit(pygame.transform.scale(img,(SW,SH)),(0,0))
    pygame.display.flip()
def pygame_pause():
    _quit = False
    while not _quit:
        for e in pygame.event.get():
            if e.type in (QUIT,MOUSEBUTTONDOWN,KEYDOWN): _quit = True
        pygame.time.wait(100)
    return

################################################################################
# UTIL FUNCTIONS
################################################################################
    
def rect_serialize(r):
    return {'x':r.x,'y':r.y,'w':r.w,'h':r.h}
def pick(d,**p):
    return p.get(MODE,d)
def save_as(fname):
    return os.path.join(DESTDIR,os.path.split(fname)[-1])
def _ppp(e,s):
    if type(e) is dict:
        for k,v in e.items(): e[k] = round(v / float(s))
    else:
        e = round(e / float(s))
    return e
        
################################################################################
# DATA TYPES
################################################################################

def img_load(v):
    if hasattr(v,'get_width'): return v
    return pygame.image.load(v)

class Atlas:
    def __init__(self,dest,src,pad=1,scale=1):
        self.dest,self.src,self.pad,self.scale = dest,src,pad,scale
        self.name=None
    def get_target(self):
        return save_as(self.dest)
    def get_deps(self):
        return self.src.values()
    def build(self):
        r = []
        pix = 0
        for name,fname in self.src.items():
            img = img_load(fname)
            rect = Rect(0,0,img.get_width(),img.get_height(),img,name)
            r.append(rect)
            pix += img.get_width()*img.get_height()
        r.sort(by_size)
        ok = False
        for size in [
            (256,256),
            (512,256),(256,512),(512,512),
            (256,1024),(1024,256),(1024,512),(512,1024),(1024,1024),
            (2048,512),(512,2048),(2048,1024),(1024,2048),(2048,2048),
            ]:
            area = size[0]*size[1]
            if pix > area: continue
            res = pack(Rect(0,0,size[0],size[1]),r[:],(0,0,0,0),self.pad)
            self._data = res
            
            tex = pygame.Surface(size).convert_alpha()
            tex.fill((0,0,0,0))
            for img in res:
                tex.blit(img.data,(img.x,img.y))
            pygame_show(tex)
            
            if len(res) != len(r):
                continue
            
            tmp = save_as(self.dest)
            self._rect = Rect(0,0,tex.get_width(),tex.get_height())
            
            if self.scale != 1:
                tex = pygame.transform.smoothscale(tex,(int(tex.get_width()/self.scale),int(tex.get_height()/self.scale)))
            pygame.image.save(tex,tmp)
            ok = True
            break
        if not ok:
            raise Exception('Atlas(%s) failed'%self.dest)
    def get_info(self):
        return {
            'fname':self.dest,
            'rect':rect_serialize(self._rect),
        }
    def get_children(self):
        r = {}
        for v in self._data:
            r[v.name] = {
                'texture':self.name,
                'rect':rect_serialize(v),
                'shape':rect_serialize(Rect(0,0,v.w,v.h)),
            }
        return r


def by_size(b,a):
    return int(max(a.w,a.h)-max(b.w,b.h))

class Rect:
    def __init__(self,x,y,w,h,data=None,name=None):
        self.x,self.y,self.w,self.h,self.data,self.name=float(x),float(y),float(w),float(h),data,name

def pack(rect,r,pad=(0,0,0,0),p=1):
    _p = 2
    return _pack(Rect(rect.x+pad[3]*_p,rect.y+pad[0]*_p,
        rect.w-(pad[1]+pad[3])*_p,rect.h-(pad[0]+pad[2])*_p),r,p)

def _pack(rect,r,p):
    # thanks to mjau for this (lua version)
    for img in r:
        if rect.w < img.w or rect.h < img.h: continue
        r.remove(img)
        img.x,img.y = rect.x,rect.y
        #  +---+----+     +---+----+
        #  |   | a1 |     |   |    |
        #  +---+----+     +---+ b2 |
        #  |   a2   |     |b1 |    |
        #  |        |     |   |    |
        #  +------- +     +---+----+
        a1 = Rect( x = rect.x + img.w, y = rect.y, w = rect.w - img.w, h = img.h )
        a2 = Rect( x = rect.x, y = rect.y + img.h, w = rect.w, h = rect.h - img.h )
        b1 = Rect( x = rect.x, y = rect.y + img.h, w = img.w, h = rect.h - img.h )
        b2 = Rect( x = rect.x + img.w, y = rect.y, w = rect.w - img.w, h = rect.h )
        
        res = [img]
        if min(a1.w, a1.h) * min(a2.w, a2.h) > min(b1.w, b1.h) * min(b2.w, b2.h):
            res.extend(pack(a1,r,(0,0,0,p),p))
            res.extend(pack(a2,r,(p,0,0,0),p))
            return res
        else:
            res.extend(pack(b1,r,(p,0,0,0),p))
            res.extend(pack(b2,r,(0,0,0,p),p))
            return res
    return []

################################################################################
class Sound:
    def __init__(self,dest,src,cmd = None):
        self.dest,self.src,self.cmd = dest,src,cmd
    def get_target(self):
        return save_as(self.dest)
    def get_deps(self):
        return [self.src]
    def build(self):
        cmd = self.cmd
        if cmd != None:
            cmd = cmd.replace('$SRC',self.src)
            cmd = cmd.replace('$DEST',save_as(self.dest))
            os.system(cmd)
            return
        shutil.copy(self.src,save_as(self.dest))
    def get_info(self):
        return { 'fname':self.dest }
    def get_children(self): return {}

################################################################################
class Font:
    def __init__(self,dest,src,size,border=0,space=0,ppp=1):
        self.dest,self.src,self.size,self.ppp = dest,src,size,ppp
        self.border,self.space = border,space
    def get_target(self):
        return save_as(self.dest)
    def get_deps(self):
        return [self.src]
    def build(self):
        f = pygame.font.Font(self.src,int(self.size))
        fg = (255,255,255)
        bg = (0,0,0)
        args = {}
        for n in xrange(0,256):
            try:
                c = chr(n)
                img1 = f.render(c,1,fg)
                img2 = f.render(c,1,bg)
            except:
                c = " "
                img1 = f.render(c,1,fg)
                img2 = img1
            # draw border stuff
            w = img1.get_width()
            if c == ' ': w += self.space
            h = img1.get_height()
            pad = self.border
            
            # HACK: fixes so that post-scale sizes are integer
            ww,hh = w+pad*2,h+pad*2
            if (ww%self.ppp) != 0: ww += self.ppp-(ww%self.ppp)
            if (hh%self.ppp) != 0: hh += self.ppp-(hh%self.ppp)
            
            img = pygame.Surface((w+pad*2,h+pad*2)).convert_alpha()
            img.fill((0,0,0,0))
            for dx in xrange(-pad,pad+1):
                for dy in xrange(-pad,pad+1):
                    img.blit(img2,(pad+dx,pad+dy))
            img.blit(img1,(pad,pad))
            args[n] = img
        self.atlas = Atlas(self.dest,args)
        self.atlas.build()
        

    def get_info(self):
        r = [None for n in range(0,256)]
        for k,e in self.atlas.get_children().items():
            r[k] = _ppp(e['rect'],self.ppp)
        return {
            'texture':self.dest,
            'rect':_ppp(self.atlas.get_info()['rect'],self.ppp),
            'height':r[0]['h'],
            'chars':r
        }
    
    def get_children(self):
        # this saves our texture info
        return {
            self.dest:{
            'fname':self.dest,
            'rect':_ppp(self.atlas.get_info()['rect'],self.ppp),
            }
        }

################################################################################
class Image:
    def __init__(self,dest,src,cmd = None):
        self.dest,self.src,self.cmd = dest,src,cmd
    def get_target(self):
        return save_as(self.dest)
    def get_deps(self):
        return [self.src]
    def build(self):
        cmd = self.cmd
        if cmd != None:
            cmd = cmd.replace('$SRC',self.src)
            cmd = cmd.replace('$DEST',save_as(self.dest))
            os.system(cmd)
        else:
            shutil.copy(self.src,save_as(self.dest))
        img = pygame.image.load(self.src)
        self.rect = pygame.Rect(0,0,img.get_width(),img.get_height())
    def get_info(self):
        return {
            'texture':self.dest,
            'rect':rect_serialize(self.rect),
            'shape':rect_serialize(Rect(0,0,self.rect.w,self.rect.h)),
        }
    def get_children(self):
        return {
            self.dest:{
            'fname':self.dest,
            'rect':rect_serialize(self.rect),
            }
        }

################################################################################
class Texture:
    def __init__(self,dest,src,cmd = None):
        self.dest,self.src,self.cmd = dest,src,cmd
    def get_target(self):
        return save_as(self.dest)
    def get_deps(self):
        return [self.src]
    def build(self):
        cmd = self.cmd
        if cmd != None:
            cmd = cmd.replace('$SRC',self.src)
            cmd = cmd.replace('$DEST',save_as(self.dest))
            os.system(cmd)
        else:
            shutil.copy(self.src,save_as(self.dest))
        img = pygame.image.load(save_as(self.dest))
        self.rect = pygame.Rect(0,0,img.get_width(),img.get_height())
    def get_info(self):
        return {
            'fname':self.dest,
            'rect':rect_serialize(self.rect),
        }
    def get_children(self): return {}
