#include "cuz_db.h"

void DBWrap::open(const char *fname) {
    int rc = sqlite3_open(fname, &this->_db);
    if( rc ){
        fprintf(stderr, "Can't open database: %s\n", sqlite3_errmsg(this->_db));
        sqlite3_close(this->_db);
        this->_db = 0;
    }
}


DBResult DBWrap::query(const char *sql, DBParams params) {
    DBResult r;
    
    this->error = 1;
        
    if (this->_db) {
        sqlite3_stmt *stmt = 0;
        int rc;
        
        // prepare the statement
        rc = sqlite3_prepare_v2(this->_db,sql,strlen(sql),&stmt,0);
        if( rc!=SQLITE_OK ) {
            fprintf(stderr, "sqlite3:prepare [%d] %s\n",sqlite3_errcode(this->_db),sqlite3_errmsg(this->_db) );
            sqlite3_finalize(stmt);
            return r;
        }
        
        // bind the variables
        if (params.size()) {
            for (int n=0; n<params.size(); n++) {
                std::string v = params[n];
                rc = sqlite3_bind_text(stmt, n+1, v.c_str(), v.size(), SQLITE_TRANSIENT);
                if (rc!=SQLITE_OK) {
                    fprintf(stderr, "sqlite3:bind [%d] %s\n",sqlite3_errcode(this->_db),sqlite3_errmsg(this->_db) );
                    sqlite3_finalize(stmt);
                    return r;
                }
            }
        }
        
        // execute the statement
        // does not handle any error situations, or busy DB or anything
        while ((rc=sqlite3_step(stmt)) == SQLITE_ROW) {
            int t = sqlite3_column_count(stmt);
            DBRow e;
            for (int i=0; i<t; i++) {
                const unsigned char *txt = sqlite3_column_text(stmt,i);
                int sz = sqlite3_column_bytes(stmt,i);
                const char *name = sqlite3_column_name(stmt,i);
                e[std::string(name)] = std::string((char*)txt,sz);
            }
            r.push_back(e);
        }
        if (rc != SQLITE_DONE) {
            fprintf(stderr, "sqlite3:step [%d] %s\n",sqlite3_errcode(this->_db),sqlite3_errmsg(this->_db) );
            sqlite3_finalize(stmt);
            return r;
        }
        
        // cleanup
        sqlite3_finalize(stmt);
    }
    
    this->error = 0;
    
    return r;
}

void DBWrap::begin() {
    this->query("begin");
}

void DBWrap::commit() {
    this->query("commit");
}

void DBWrap::close() {
    if (!this->_db) { return; }
    sqlite3_close(this->_db);
    this->_db = 0;
}

