#include "driver-ios.h"

#import <OpenGLES/ES1/gl.h>
#import <OpenGLES/ES1/glext.h>

#include <AudioToolbox/AudioToolbox.h>
#include <CoreFoundation/CoreFoundation.h>
#include <AVFoundation/AVFoundation.h>

#include "Texture2D.h"
#include "SoundEngine.h"

AVAudioPlayer *m_bgMusicPlayer = 0;
Texture2D *texs[256];
int texs_init = 0;
int _driver_keyboard_state = 0;
IBOutlet UITextField *tfield;
int _driver_keyboard_height = 0;

const char *_driver_get_docdir_ios() {
    static char s[512];
    NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
    NSString *documentsDirectory = [paths objectAtIndex:0];
    sprintf(s,"%s",[documentsDirectory UTF8String]);
    return s;
}


void _driver_audio_init() {
    // Set the sound priority such that it can mix with music (my music, or music running in the background by a user)
    AudioSessionInitialize(NULL, NULL, NULL, NULL);
	UInt32 sessionCategory = kAudioSessionCategory_AmbientSound; //kAudioSessionCategory_UserInterfaceSoundEffects;//kAudioSessionCategory_AmbientSound; // you could also use kAudioSessionCategory_UserInterfaceSoundEffects
	AudioSessionSetProperty(kAudioSessionProperty_AudioCategory, sizeof(sessionCategory), &sessionCategory);
	SoundEngine_Initialize( 0 );
	SoundEngine_SetListenerPosition(0,0,0);
}

void _driver_audio_deinit() {
	SoundEngine_Teardown();
}

int driver_audio_sample_load(const char *fname) {
    fprintf(stderr,"driver_audio_sample_load:ios(%s)\n",fname);
    UInt32 n;
    SoundEngine_LoadEffect(fname, &n);
    return (int)n;
}


void driver_audio_sample_play(int n, float vol=1.0, float pan=0.0) {
    SoundEngine_SetEffectLevel(n,vol);
    SoundEngine_SetEffectPosition(n,  pan,0,0);
    SoundEngine_StartEffect( n );
}

void driver_audio_sample_free(int n) {
    SoundEngine_UnloadEffect( n );
}

int driver_audio_stream_load(const char *fname, bool preload) {
    // check if we've got music playing already ...
    UInt32 isPlaying;
    UInt32 propertySize = sizeof(isPlaying);
    AudioSessionGetProperty(kAudioSessionProperty_OtherAudioIsPlaying, &propertySize, &isPlaying);
    if (isPlaying) { return 0; }
    
    // try to play some music
    NSString *soundFile =  [NSString stringWithCString: fname encoding: [NSString defaultCStringEncoding]];

    if (m_bgMusicPlayer) {
        [m_bgMusicPlayer stop];
        [m_bgMusicPlayer release];
        m_bgMusicPlayer = NULL;
    }
    
    m_bgMusicPlayer = [AVAudioPlayer alloc ];
    [m_bgMusicPlayer initWithContentsOfURL: [NSURL fileURLWithPath:soundFile] error:nil];
    m_bgMusicPlayer.numberOfLoops = -1;

    [m_bgMusicPlayer prepareToPlay];
    return 1;
}

void driver_audio_stream_play(int n, float v, float pos) {
    if (m_bgMusicPlayer) {
        [m_bgMusicPlayer play];
        m_bgMusicPlayer.volume = v;
    }
}

void driver_audio_stream_volume(int n, float v) {
    if (m_bgMusicPlayer) {
        m_bgMusicPlayer.volume = v;
    }
}

void driver_audio_stream_stop(int n) {
}

void driver_audio_stream_free(int n) {

}



int _driver_tex_load_ios(const char *fname) {
    int _ww,_hh; int *ww = &_ww; int *hh = &_hh;
    
    if (!texs_init) {
        memset(texs,0,sizeof(texs));
    }
    printf("\ndriver_tex_load_ios(%s)\n",fname);
    NSString *s = [[NSString alloc] initWithUTF8String: fname];

    Texture2DPixelFormat format =  kTexture2DPixelFormat_RGBA8888;
    
//    if (strstr(fname,"font")) {
//        printf("font -> format:RGBA4444\n");
//        format = kTexture2DPixelFormat_RGBA4444;
//    }
//    if (strstr(fname,".jpg")!=0) {
//        printf("jpg -> format:RGB565\n");
//        format = kTexture2DPixelFormat_RGB565;
//    }
    
    [Texture2D setDefaultAlphaPixelFormat:format];
    
    Texture2D *tex = [[Texture2D alloc] initWithImage:[UIImage imageWithContentsOfFile :s]];

    int r = [tex name];
    *ww = [tex pixelsWide];
    *hh = [tex pixelsHigh];
    
    texs[texs_init++] = tex;
        
    [tex setAntiAliasTexParameters];
    glBindTexture( GL_TEXTURE_2D, r );
//    if (*ww == *hh && strstr(fname,"background") == 0) {
//        printf("enabling mipmaps\n");
//        glGenerateMipmapOES(GL_TEXTURE_2D);
//        _ccTexParams texParams = { GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE };
//        [tex setTexParameters:&texParams];
//    }
        
    [s release];
    printf("loaded %s as %d : %d x %d\n",fname,r,*ww,*hh);
    return r;
}


void driver_keyboard_init(void *_view) {
    UIView<UITextFieldDelegate> *view = (UIView *)_view;
    CGRect frame = CGRectMake(0,0,1024,21);
    tfield = [[UITextField alloc] initWithFrame:frame];
    tfield.delegate = view;
    [view addSubview: tfield];
    
    // styling: http://developer.apple.com/iphone/library/documentation/UIKit/Reference/UITextInputTraits_Protocol/Reference/UITextInputTraits.html
    
    tfield.autocapitalizationType = UITextAutocapitalizationTypeNone;
    tfield.autocorrectionType = UITextAutocorrectionTypeNo;
    tfield.enablesReturnKeyAutomatically = NO;
    tfield.keyboardAppearance = UIKeyboardAppearanceDefault;
    //tfield.keyboardType = UIKeyboardTypeDefault;
    tfield.keyboardType = UIKeyboardTypeASCIICapable;
    tfield.returnKeyType = UIReturnKeyDefault;
    //tfield.returnKeyType = UIReturnKeyDone;
    tfield.secureTextEntry = NO;
    
    tfield.placeholder = @""; // "type something here"
    tfield.text = @"  "; // starting text, all are 2 spaces for consistency ..
    
    // visuals for testing
    tfield.textColor = [UIColor blackColor];
    tfield.backgroundColor = [UIColor whiteColor];
    tfield.hidden = YES;
    
}

void driver_keyboard_show() {
	tfield.text = @"  ";
    [tfield becomeFirstResponder];
	_driver_keyboard_state = 1;
	fprintf(stderr,"keyboard show\n");
}

void driver_keyboard_hide() {
    [tfield resignFirstResponder];
	_driver_keyboard_state=0;
	fprintf(stderr,"keyboard hide\n");
}

int driver_keyboard_state() { return _driver_keyboard_state; }
void driver_keyboard_set_height(int v) { _driver_keyboard_height = v; }
int driver_keyboard_get_height() { return _driver_keyboard_height * _driver_keyboard_state; }

void _driver_open_url_ios(const char *url) {
    NSString *myurl = [[NSString alloc] initWithUTF8String: url];
    [[UIApplication sharedApplication] openURL:[NSURL URLWithString:myurl]];
}

void _driver_video_init() { }
void driver_video_flip() { }
long driver_timer_get_ticks() { return 0; }
void _driver_video_deinit() { }
void driver_video_init(int ww, int hh, int dw, int dh, int mw, int mh, int full) { }
void driver_timer_delay(int v) { }
bool driver_video_minimize() { return false; }
