////////////////////////////////////////////////////////////////////////////////
#ifdef USING_IRRKLANG

#include "cuz_driver.h"

#include <sys/types.h>
#include <sys/stat.h>
#ifndef PLATFORM_WIN32
#include <unistd.h>
#endif

#ifdef PLATFORM_WIN32
#pragma comment(lib, "../../cuzcode/ext/irrklang/windows/irrKlang.lib")
#define IRRKLANG_STATIC
#endif

#include "irrKlang.h"
using namespace irrklang;

long _driver_audio_get_ticks() { return driver_timer_get_ticks(); }

struct _DriverAudioData {
    ISoundSource *samples[256]; int samples_n;
    ISound *channels[256][2];
    ISound *streams[256]; int streams_n;
    ISoundEngine* engine;
    
    _DriverAudioData() {
        memset(this,0,sizeof(_DriverAudioData));
        samples_n=1;
        streams_n=1;
    }
};
static _DriverAudioData *_driver_audio = 0;

bool _driver_audio_is_init() {
    if (!_driver_audio) { return false; }
    if (!_driver_audio->engine) { return false; }
    return true;
}

void _driver_audio_init() {
    if (_driver_audio) { return; }
    _driver_audio = new _DriverAudioData();
    _driver_audio->engine = createIrrKlangDevice();
    if (!_driver_audio_is_init()) {
        fprintf(stderr,"IrrKlang init error\n");
    }
}

void _driver_audio_deinit() {
    if (!_driver_audio) { return; }
    if (_driver_audio_is_init()) {
        _driver_audio->engine->drop();
        _driver_audio->engine = 0;
    }
    delete _driver_audio;
    _driver_audio = 0;
}

int driver_audio_sample_load(const char *fname) {
    if (!_driver_audio_is_init()) { return 0; }
    
    //const char *fname = driver_data_fname(_fname);
    fprintf(stderr,"loading: %s\n",fname);
    int &n = _driver_audio->samples_n;
    _driver_audio->samples[n] = _driver_audio->engine->addSoundSourceFromFile(fname);
    if (!_driver_audio->samples[n]) {
        fprintf(stderr,"error loading: %s\n",fname);
        return 0;
    }
    return n++;
}


void driver_audio_sample_play(int sid, float vol, float pan) {
    if (sid <= 0) { return; }
    if (!_driver_audio_is_init()) { return; }
    if (!_driver_audio->samples[sid]) { fprintf(stderr,"sound_play error: %d - missing sample\n",sid); return; }
    
    ISound *tmp;
    
    if ((tmp=_driver_audio->channels[sid][0])!=0) { tmp->stop(); _driver_audio->channels[sid][0] = 0; }
    if ((tmp=_driver_audio->channels[sid][1])!=0) { tmp->stop(); _driver_audio->channels[sid][1] = 0; }
    
    pan = (pan + 1)/2.0;
    
    ISound *snd_left = _driver_audio->engine->play2D(_driver_audio->samples[sid],false,true,true,false);
    if (snd_left) {
        snd_left->setPan(0);
        snd_left->setVolume(vol*(1.0-pan));
    }
    ISound *snd_right = _driver_audio->engine->play2D(_driver_audio->samples[sid],false,true,true,false);
    if (snd_right) {
        snd_right->setPan(-1);
        snd_right->setVolume(vol*pan);
    }
    if (snd_left) { snd_left->setIsPaused(false); }
    if (snd_right) { snd_right->setIsPaused(false); }
    
    _driver_audio->channels[sid][0] = snd_left;
    _driver_audio->channels[sid][1] = snd_right;
}


int driver_audio_stream_load(const char *fname, bool preload) {
    if (!_driver_audio_is_init()) { return 0; }
    //const char *fname = driver_data_fname(_fname);
    int &n = _driver_audio->samples_n;
    
    int tt = _driver_audio_get_ticks();
    
    ISoundSource *ss = 0;
    
    if (!preload) {
        ss = _driver_audio->engine->addSoundSourceFromFile(fname);
    } else {
        struct stat stbuf;
        stat(fname, &stbuf);
        int fsize = stbuf.st_size;
        FILE* f = fopen(fname,"rb");
        if (!f) { fprintf(stderr,"stream_load fopen error: %s\n",fname); return n++; }
        if (!fsize) { fprintf(stderr,"stream_load, fsize=0: %s\n",fname); return n++; }
        char* mem = new char[fsize];
        fread(mem, fsize, 1, f);
        fclose(f);
        ss = _driver_audio->engine->addSoundSourceFromMemory(mem, fsize, fname, true);
        delete [] mem;
    }
    
    if (ss != 0) {
        // this bytes value seems to be the uncompressed in-memory length
        // of the music ..
        //         ss->setForcedStreamingThreshold(32*1024*1024);
        //         ss->setStreamMode(preload?ESM_NO_STREAMING:ESM_AUTO_DETECT);
        ss->setStreamMode(ESM_AUTO_DETECT);
        _driver_audio->streams[n] = _driver_audio->engine->play2D(ss,true,true,true,false);
        ss->drop();
        if (_driver_audio->streams[n] == 0) {
            fprintf(stderr,"stream_load (play2d) error: %s\n",fname);
        }
    } else {
        fprintf(stderr,"stream_load (soundSource) error: %s\n",fname);
    }
    
    fprintf(stderr,"stream_load in ticks: %d\n",_driver_audio_get_ticks()-tt);
    
    return n++;
}

float driver_audio_stream_len(int sid) {
    if (sid <= 0) { return 1; }
    if (!_driver_audio_is_init()) { return 1; }
    if (!_driver_audio->streams[sid]) { fprintf(stderr,"stream_len error: %d - missing stream\n",sid); return 1; }
    
    int len = _driver_audio->streams[sid]->getPlayLength();
    return len/1000.0;
}

void driver_audio_stream_play(int sid, float vol, float pos) {
    if (sid <= 0) { return ; }
    if (!_driver_audio_is_init()) { return ; }
    if (!_driver_audio->streams[sid]) { fprintf(stderr,"stream_play error: %d - missing stream\n",sid); return ; }
    
    int tt = _driver_audio_get_ticks();
    
    _driver_audio->streams[sid]->setVolume(vol);
    fprintf(stderr,"setvol: %d\n",_driver_audio_get_ticks()-tt); tt = _driver_audio_get_ticks();
    _driver_audio->streams[sid]->setPlayPosition((int)(pos*1000));
    fprintf(stderr,"setplayposition: %d\n",_driver_audio_get_ticks()-tt); tt = _driver_audio_get_ticks();
    _driver_audio->streams[sid]->setIsPaused(false);
    fprintf(stderr,"setispaused: %d\n",_driver_audio_get_ticks()-tt); tt = _driver_audio_get_ticks();
    
    fprintf(stderr,"Pos: set: %d get: %d\n",(int)(pos*1000),_driver_audio->streams[sid]->getPlayPosition());
}

void driver_audio_stream_stop(int sid) {
    if (sid <= 0) { return ; }
    if (!_driver_audio_is_init()) { return ; }
    if (!_driver_audio->streams[sid]) { fprintf(stderr,"stream_stop error: %d - missing stream\n",sid); return ; }
    
    _driver_audio->streams[sid]->setIsPaused(true);
}

void driver_audio_stream_volume(int sid, float vol) {
    if (sid <= 0) { return ; }
    if (!_driver_audio_is_init()) { return ; }
    if (!_driver_audio->streams[sid]) { fprintf(stderr,"stream_volume error: %d - missing stream\n",sid); return ; }
    
    _driver_audio->streams[sid]->setVolume(vol);
}


#else

//void _driver_audio_init() { }
//
//void _driver_audio_deinit() { }
#endif

