#ifdef USING_SDL

#include "cuz_driver.h"
#include "cuz_opengl.h"
#include "cuz_util.h"

#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>
#ifndef PLATFORM_WIN32
#include <unistd.h>
#endif

#ifdef PLATFORM_WIN32
    #include <windows.h>
    #include <conio.h>
    #include <shellapi.h>
	#include <direct.h>
#endif

void _driver_video_init() {
    SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER);
    
    const SDL_version* v; SDL_version v2;
    
    v = SDL_Linked_Version();
    printf("SDL Runtime Version: %u.%u.%u\n", v->major, v->minor, v->patch);
    SDL_VERSION(&v2); v = &v2; 
    printf("SDL Compile Version: %u.%u.%u\n", v->major, v->minor, v->patch);
    
    v = IMG_Linked_Version();
    printf("SDL_image Runtime Version: %u.%u.%u\n", v->major, v->minor, v->patch);
    SDL_IMAGE_VERSION(&v2); v = &v2; 
    printf("SDL_image Compile Version: %u.%u.%u\n", v->major, v->minor, v->patch);
    
    #if SDL_BYTEORDER == SDL_LIL_ENDIAN
        printf("SDL_BYTEORDER == SDL_LIL_ENDIAN\n");
    #else
        printf("SDL_BYTEORDER == SDL_BIG_ENDIAN\n");
    #endif
}


SDL_Surface *_try_init_video(int ww, int hh, int full) {
    fprintf(stderr,"_try_init_video(%d,%d,%d)\n",ww,hh,full);
    
    //full = 0;
    
    SDL_QuitSubSystem(SDL_INIT_VIDEO);
    // have to quit and re-init so that OS/X doesn't
    // set auto-mode to the same as the previous window size
    SDL_InitSubSystem(SDL_INIT_VIDEO);

    SDL_EnableUNICODE(1);
    SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY,SDL_DEFAULT_REPEAT_INTERVAL);

    if (strlen(driver_get_caption())) {
        SDL_WM_SetCaption(driver_get_caption(),driver_get_caption());
    }
    
/*    #ifndef BUILD_OSX
    char s[512];
    sprintf(s,"%s/icon32.bmp",DRIVER_DATADIR);
	fprintf(stderr,"Loadbmp: %s\n",s);
    SDL_Surface *icon = SDL_LoadBMP(s);
	if (icon) {
		Uint32 colorkey = SDL_MapRGB(icon->format, 255, 0, 0);
		SDL_SetColorKey(icon, SDL_SRCCOLORKEY, colorkey);  
		SDL_WM_SetIcon(icon,0);
		SDL_FreeSurface(icon);
	} else {
		fprintf(stderr,"icon load fail\n");
	}
    #endif*/
    
    int flags =SDL_OPENGL | ((full!=0)*SDL_FULLSCREEN);
    
	
	int bits = 0; // however, we set it to 0, so that things "just work" when we don't specify any GL attributes
	
	const SDL_VideoInfo* VideoInfo = SDL_GetVideoInfo();
	fprintf(stderr,"VideoInfo->vfmt->BitsPerPixel:%d\n",VideoInfo->vfmt->BitsPerPixel);
	
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,16);
    
	// note: conversion persists in windowed mode though
	SDL_Surface *res = SDL_SetVideoMode(ww,hh, bits, flags);
    
    
    fprintf(stderr,"res: %d\n",res!=0);
    
    if (res) {
        int val;
        SDL_GL_GetAttribute(SDL_GL_RED_SIZE,&val);
        fprintf(stderr,"GL_RED_SIZE:%d\n",val);
        SDL_GL_GetAttribute(SDL_GL_GREEN_SIZE,&val);
        fprintf(stderr,"GL_GREEN_SIZE:%d\n",val);
        SDL_GL_GetAttribute(SDL_GL_BLUE_SIZE,&val);
        fprintf(stderr,"GL_BLUE_SIZE:%d\n",val);
        SDL_GL_GetAttribute(SDL_GL_ALPHA_SIZE,&val);
        fprintf(stderr,"GL_ALPHA_SIZE:%d\n",val);
        SDL_GL_GetAttribute(SDL_GL_DEPTH_SIZE,&val);
        fprintf(stderr,"GL_DEPTH_SIZE:%d\n",val); 
        SDL_GL_GetAttribute(SDL_GL_BUFFER_SIZE,&val);
        fprintf(stderr,"GL_BUFFER_SIZE:%d\n",val);
        SDL_GL_GetAttribute(SDL_GL_DOUBLEBUFFER,&val);
        fprintf(stderr,"GL_DOUBLEBUFFER:%d\n",val);
    }
    return res;
}


void driver_video_init(int ww, int hh, int dw, int dh, int mw, int mh, int full) {
    SDL_Surface *res;
    const SDL_VideoInfo *vinfo;
    
    fprintf(stderr,"init_video: %d %d %d %d %d %d %d\n",ww,hh,dw,dh,mw,mh,full);
    
    res = _try_init_video(ww,hh,full);
    
    if (!res) {
        fprintf(stderr,"trying at default res ...\n");
        ww = dw; hh = dh;
        res = _try_init_video(ww,hh,full);
    }
    
    if (!res) {
        fprintf(stderr,"trying w/o fullscreen ...\n");
        ww = dw; hh = dh; full = 0;
        res = _try_init_video(ww,hh,full);
    }
    
    if (res && (res->w < mw || res->h < mh)) {
        fprintf(stderr,"trying at minimum-sized windowed ...\n");
        ww = mw; hh = mh; full = 0;
        res = _try_init_video(ww,hh,full);
    }
        
    
    if (!res) { return; } // fail
    driver_set_size(res->w,res->h);
    driver_set_aspect(res->w / res->h);
//    driver_set_orient(0);
    
    // dump out our mode info
    if (res) { fprintf(stderr,"res: %d x %d\n",res->w,res->h); }
    vinfo = SDL_GetVideoInfo();
    if (vinfo) { fprintf(stderr,"vinfo: %d x %d\n",vinfo->current_w,vinfo->current_h); }
    
    
    // we picked default, so we're okay
    if (ww == 0 && hh == 0) { return; }
    
    // check if we got the mode we requested
    if (ww == res->w && hh == res->h) { return; }
    
    // we're not in the mode we asked for, so we need to
    // reset ONE more time, to appease SDL
    // otherwise it puts a bounding box around the mouse
    // that locks the mouse within your requested bounds
    ww = res->w; hh = res->h;
    res = _try_init_video(ww,hh,full);
    
    if (!res) { return; } // fail
    
    driver_set_size(res->w,res->h);
    driver_set_aspect(res->w / res->h);
//    driver_set_orient(0);
    
    if (res) { fprintf(stderr,"res: %d x %d\n",res->w,res->h); }
    vinfo = SDL_GetVideoInfo();
    if (vinfo) { fprintf(stderr,"vinfo: %d x %d\n",vinfo->current_w,vinfo->current_h); }
}

bool driver_video_minimize() {
    return SDL_WM_IconifyWindow();
}


int _driver_sdl_sym2key(int sym) {
    int v = KEY_NONE;
    
    switch (sym) {
        case SDLK_KP0: v = KEY_P0; break;
        case SDLK_KP1: v = KEY_P1; break;
        case SDLK_KP2: v = KEY_P2; break;
        case SDLK_KP3: v = KEY_P3; break;
        case SDLK_KP4: v = KEY_P4; break;
        case SDLK_KP5: v = KEY_P5; break;
        case SDLK_KP6: v = KEY_P6; break;
        case SDLK_KP7: v = KEY_P7; break;
        case SDLK_KP8: v = KEY_P8; break;
        case SDLK_KP9: v = KEY_P9; break;
        case SDLK_UP: v = KEY_UP; break;
        case SDLK_DOWN: v = KEY_DOWN; break;
        case SDLK_LEFT: v = KEY_LEFT; break;
        case SDLK_RIGHT: v = KEY_RIGHT; break;
        case SDLK_BACKSPACE: v = KEY_BACKSPACE; break;
        case SDLK_DELETE: v = KEY_DELETE; break;
        case SDLK_ESCAPE: v = KEY_ESCAPE; break;
        case SDLK_RETURN: v = KEY_RETURN; break;
        case SDLK_KP_ENTER: v = KEY_RETURN; break;
        case SDLK_TAB: v = KEY_TAB; break;
        case SDLK_F1: v = KEY_F1; break;
        case SDLK_F2: v = KEY_F2; break;
        case SDLK_F3: v = KEY_F3; break;
        case SDLK_F4: v = KEY_F4; break;
        case SDLK_F5: v = KEY_F5; break;
        case SDLK_F6: v = KEY_F6; break;
        case SDLK_F7: v = KEY_F7; break;
        case SDLK_F8: v = KEY_F8; break;
        case SDLK_F9: v = KEY_F9; break;
        case SDLK_F10: v = KEY_F10; break;
        case SDLK_F11: v = KEY_F11; break;
        case SDLK_F12: v = KEY_F12; break;
        case SDLK_a: v = KEY_a; break;
        case SDLK_b: v = KEY_b; break;
        case SDLK_c: v = KEY_c; break;
        case SDLK_d: v = KEY_d; break;
        case SDLK_e: v = KEY_e; break;
        case SDLK_f: v = KEY_f; break;
        case SDLK_g: v = KEY_g; break;
        case SDLK_h: v = KEY_h; break;
        case SDLK_i: v = KEY_i; break;
        case SDLK_j: v = KEY_j; break;
        case SDLK_k: v = KEY_k; break;
        case SDLK_l: v = KEY_l; break;
        case SDLK_m: v = KEY_m; break;
        case SDLK_n: v = KEY_n; break;
        case SDLK_o: v = KEY_o; break;
        case SDLK_p: v = KEY_p; break;
        case SDLK_q: v = KEY_q; break;
        case SDLK_r: v = KEY_r; break;
        case SDLK_s: v = KEY_s; break;
        case SDLK_t: v = KEY_t; break;
        case SDLK_u: v = KEY_u; break;
        case SDLK_v: v = KEY_v; break;
        case SDLK_w: v = KEY_w; break;
        case SDLK_x: v = KEY_x; break;
        case SDLK_y: v = KEY_y; break;
        case SDLK_z: v = KEY_z; break;
        case SDLK_HOME: v = KEY_HOME; break;
        case SDLK_END: v = KEY_END; break;
    }
    return v;
}

int driver_event_poll(Event *r) {
    SDL_Event e;
    memset(&e,0,sizeof(e));
    memset(r,0,sizeof(Event));
    r->type = EVT_NONE;

    if (!SDL_PollEvent(&e)) { return 0; }
    
    switch (e.type) {
        case SDL_ACTIVEEVENT:
            r->type = EVT_SDLEVENT;
            r->stype = e.type;
            r->b = e.active.state;
            r->v = e.active.gain;
            //fprintf(stderr,"sdl_activeevent: %d %d\n",r->b,r->v);
            break;
        case SDL_QUIT:
            r->type = EVT_QUIT;
            break;
        case SDL_MOUSEMOTION:
            #ifdef BUILD_IPHONE_EMULATE
            if (e.motion.state != 0) {
            #endif
                r->type = EVT_MOTION;
                r->_x = e.motion.x;
                r->_y = e.motion.y;
                r->_dx = e.motion.xrel;
                r->_dy = e.motion.yrel;
                r->b = e.motion.state;
				r->device = e.motion.which;
            #ifdef BUILD_IPHONE_EMULATE
            }
            #endif
            break;
        case SDL_MOUSEBUTTONDOWN:
            r->type = EVT_DOWN;
            r->_x = e.button.x;
            r->_y = e.button.y;
            r->b = e.button.button;
			r->device = e.button.which;
//             fprintf(stderr,"mousedown: %d %d %d\n",e.button.x,e.button.y,e.button.button);
            if (e.button.button == SDL_BUTTON_WHEELUP || e.button.button == SDL_BUTTON_WHEELDOWN) {
                r->type = EVT_WHEEL;
                r->_dy = (e.button.button==SDL_BUTTON_WHEELUP?-1:1);
            }
            break;
        case SDL_MOUSEBUTTONUP:
            r->type = EVT_UP;
            r->_x = e.button.x;
            r->_y = e.button.y;
            r->b = e.button.button;
			r->device = e.button.which;
//             fprintf(stderr,"mouseup: %d %d %d\n",e.button.x,e.button.y,e.button.button);
            if (e.button.button == SDL_BUTTON_WHEELUP || e.button.button == SDL_BUTTON_WHEELDOWN) {
                r->type = EVT_NONE;
            }
            break;
            
        case SDL_KEYUP:
            r->type = EVT_KEYUP;
            r->v = _driver_sdl_sym2key(e.key.keysym.sym);
            r->c = e.key.keysym.unicode;
            r->b = 0;
            r->b |= KEYM_ALT*((KMOD_LALT&e.key.keysym.mod)!=0);
            r->b |= KEYM_ALT*((KMOD_RALT&e.key.keysym.mod)!=0);
            r->b |= KEYM_SHIFT*((KMOD_LSHIFT&e.key.keysym.mod)!=0);
            r->b |= KEYM_SHIFT*((KMOD_RSHIFT&e.key.keysym.mod)!=0);
            r->b |= KEYM_CTRL*((KMOD_LCTRL&e.key.keysym.mod)!=0);
            r->b |= KEYM_CTRL*((KMOD_RCTRL&e.key.keysym.mod)!=0);
            r->b |= KEYM_OPT*((KMOD_LMETA&e.key.keysym.mod)!=0);
            r->b |= KEYM_OPT*((KMOD_RMETA&e.key.keysym.mod)!=0);
            break;
            
        case SDL_KEYDOWN:
            r->type = EVT_KEYDOWN;
            r->v = _driver_sdl_sym2key(e.key.keysym.sym);
            r->c = e.key.keysym.unicode;
            r->b = 0;
            r->b |= KEYM_ALT*((KMOD_LALT&e.key.keysym.mod)!=0);
            r->b |= KEYM_ALT*((KMOD_RALT&e.key.keysym.mod)!=0);
            r->b |= KEYM_SHIFT*((KMOD_LSHIFT&e.key.keysym.mod)!=0);
            r->b |= KEYM_SHIFT*((KMOD_RSHIFT&e.key.keysym.mod)!=0);
            r->b |= KEYM_CTRL*((KMOD_LCTRL&e.key.keysym.mod)!=0);
            r->b |= KEYM_CTRL*((KMOD_RCTRL&e.key.keysym.mod)!=0);
            r->b |= KEYM_OPT*((KMOD_LMETA&e.key.keysym.mod)!=0);
            r->b |= KEYM_OPT*((KMOD_RMETA&e.key.keysym.mod)!=0);
//    		fprintf(stderr,"keydown: %d, %d, %d, %d, %d, %d\n",(int)r->type,(int)sym,(int)r->c, (int)r->v,(int)SDLK_ESCAPE,(int)KEY_ESCAPE);
            break;
    }
    r->x = r->_x; r->dx = r->_dx; r->y = r->_y; r->dy = r->_dy;

    return 1;

}

void driver_video_flip() {
    SDL_GL_SwapBuffers();
}

void _driver_video_deinit() {
    SDL_Quit();
}


int _driver_tex_load_sdl(const char *fname) {
    
    // load the surface, then convert it to the correct format
    SDL_Surface* _surf = IMG_Load(fname);
    if (!_surf) {
        fprintf(stderr,"unable to load %s\n",fname);
        return 0;
    }
    SDL_Surface *surface;
    SDL_PixelFormat fmt;
    fmt.palette = 0;
    fmt.BitsPerPixel = 32;
    fmt.BytesPerPixel = 4;
    fmt.Rloss=0; fmt.Gloss=0; fmt.Bloss=0; fmt.Aloss=0;
    #if SDL_BYTEORDER == SDL_LIL_ENDIAN
    fmt.Rshift = 0; fmt.Gshift = 8; fmt.Bshift = 16; fmt.Ashift = 24;
    fmt.Rmask = 0x000000ff; fmt.Gmask = 0x0000ff00;
    fmt.Bmask = 0x00ff0000; fmt.Amask = 0xff000000;
    #else
    fmt.Rshift = 24; fmt.Gshift = 16; fmt.Bshift = 8; fmt.Ashift = 0;
    fmt.Rmask = 0xff000000; fmt.Gmask = 0x00ff0000;
    fmt.Bmask = 0x0000ff00; fmt.Amask = 0x000000ff;
    #endif
    fmt.colorkey = 0;
    fmt.alpha = 255;
    surface = SDL_ConvertSurface(_surf, &fmt,SDL_SRCALPHA);
    SDL_FreeSurface(_surf);

    SDL_PixelFormat *format = surface->format;
    int x,y;
    // filter out bad pixel colors
    if (format->Amask != 0 && format->BytesPerPixel == 4) {
        Uint32 *pix = (Uint32*)surface->pixels;
        int pitch = surface->pitch/4;
        for (y=0; y<surface->h; y++) {
            for (x=0; x<pitch; x++) {
                if (((*pix)&format->Amask)==0) { *pix=0; }
                pix++;
            }
        }
    } else {
        fprintf(stderr,"surf_to_tex2: no Amask\n");
    }
	
    GLuint texture;
    glPixelStorei(GL_UNPACK_ALIGNMENT,4);
    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_2D, texture);
    
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_LINEAR);
//     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
    
    gluBuild2DMipmaps(GL_TEXTURE_2D, 4,
        surface->w, surface->h, GL_RGBA,GL_UNSIGNED_BYTE, surface->pixels);
    SDL_FreeSurface(surface);

    fprintf(stderr,"loaded %s as %d\n",fname,texture);
    return texture;
}


long driver_timer_get_ticks() {
    return SDL_GetTicks();
}

void driver_timer_delay(int ticks) {
    SDL_Delay(ticks);
}



void driver_keyboard_init(void *_view) {}
void driver_keyboard_show() {}
void driver_keyboard_hide() {}
int driver_keyboard_state() {return 0;}
void driver_keyboard_set_height(int v) {}
int driver_keyboard_get_height() {return 0;}

#endif
