#ifndef CUZ_DRIVER_H
#define CUZ_DRIVER_H

#include "cuz_event.h"

#include <string>

void driver_init();

void driver_video_init(int ww, int hh, int dw, int dh, int mw, int mh, int full);

bool driver_video_minimize(); // returns false on failure


int driver_event_poll(Event *e);

// virtual size
//void driver_set_vsize(int w, int h);
//void driver_get_vsize(int *w, int *h);
//void driver_set_vorient(float a);

// physical size
void driver_set_size(int w, int h);
void driver_get_size(int *w, int *h);
void driver_set_aspect(float a);
float driver_get_aspect();

// handy functions!
//float driver_get_rotation() ; // find out (given the orient, vorient, how much to rotate the screen to render

// take physical Events and translate them to our virtual coordinates
//void driver_event_correct(Event *e);

void driver_video_flip();

// void driver_video_get_size(int *w, int *h);

void driver_deinit();

int driver_tex_load(const char *fname);

std::string driver_data_load(const char *fname);
const char *driver_data_fname(const char *fname);

void driver_set_gamename(const char *name);
void driver_set_caption(const char *v);
const char *driver_get_caption();
void driver_doc_save(const char *fname, void *data, int len);
int driver_doc_load(const char *fname, void *data, int len);
//std::string driver_doc_load(const char *fname);
const char *driver_doc_fname(const char *fname);

void driver_timer_set(int fps);
void driver_timer_tick();
long driver_timer_get_ticks();
void driver_timer_delay(int ticks);

void driver_open_url(const char *s) ;


int driver_audio_sample_load(const char *s);
void driver_audio_sample_play(int v, float vol=1.0, float pan = 0.0);

int driver_audio_stream_load(const char *s, bool preload = false);
void driver_audio_stream_play(int v, float vol = 1.0, float pos = 0.0);
void driver_audio_stream_stop(int v);
void driver_audio_stream_volume(int v, float vol);

void driver_clipboard_set_text(const std::string &text);
std::string driver_clipboard_get_text();

// virtual keyboard controls
void driver_keyboard_init(void *_view) ;
void driver_keyboard_show() ;
void driver_keyboard_hide() ;
int driver_keyboard_state() ;
void driver_keyboard_set_height(int v) ;
int driver_keyboard_get_height();

#endif
