#include <stdio.h>

#include "cuz_driver.h"
#include "cuz_font.h"
#include "cuz_gfx.h"

#include "cuz_opengl.h"

#include "cuz_json.h"
#include <map>

#define FONT_LINES_MAX 32
struct FontWrap {
    int total;
    int lines[FONT_LINES_MAX][3];
};

struct _FontData {
    var json;
    std::map<std::string,int> map;
    FontInfo data[256];
    int n;
    
    _FontData() {
        memset(&data,0,sizeof(data));
        n = 1;
    }
};

static _FontData *_font = 0;


void font_init() {
    _font = new _FontData();

    std::string json;
     
    json = driver_data_load("data.json");
    json_decode(_font->json,json.c_str(),json.size());
}

void font_deinit() {
    delete _font;
    _font = 0;
}

extern GfxRect _gfx_json_rect(var &js);

int font_load(const char *name) {
    var &js = _font->json;
    
    if (_font->map.find(name) == _font->map.end()) {
        fprintf(stderr,"trying load font %s\n",name);
        if (!js.has(name)) {
            fprintf(stderr,"unable to find font %s\n",name);
            _font->map[name] = 0;
            return 0;
        }
        int n = _font->n;
        fprintf(stderr,"doing load font %s (%d)\n",name,n);
        _font->map[name] = n;
        _font->data[n].n = n;
        _font->data[n].height = js[name]["height"];
        _font->data[n].tex = gfx_tex_load(js[name]["texture"].c_str());
        _font->data[n].rect = _gfx_json_rect(js[name]["rect"]);
        int i; for (i=0;i<256;i++) {
            _font->data[n].chars[i] = _gfx_json_rect(js[name]["chars"][i]);
        }
        _font->n += 1;
    }
    
    return _font->map[name];
}


#define _FONT_FILTER_MAX 1024
    
const char *_font_filter(const char *s) {
    static char s3[_FONT_FILTER_MAX];
    static char s2[_FONT_FILTER_MAX];
    memset(s2,0,sizeof(s2));
    memset(s3,0,sizeof(s3));
    int j,l=strlen(s);
    if (l>=_FONT_FILTER_MAX) { l = _FONT_FILTER_MAX-1; }
    for (j=0; j<l;j++) { s2[j]=tolower(s[j]); }
    strcpy(s3,s);
    
    const char *badwords[] = {
    "bastard", "bitch","clit", "cunt", "damn","dick", "fag", "fuck", "nigger", "penis",
    "piss", "pussy", "shit","fck","fuk","phuck","wtf","phuk",
    "asshole","cuunt","f@g","fuuck","vagina","smeg","klit","kunt","wank",
    "fock",
    0 };
    
    int i; const char *word; const char *pos;
    for (i=0; (word=badwords[i])!=0; i++) {
        while ((pos=strstr(s2,word))!=0) {
            for (j=0; j<(int)strlen(word); j++) {
                s2[j+pos-s2] = '*';
                s3[j+pos-s2] = '*';
            }
        }
    }

    return s3;
}
    
    
bool _gfx_clip(GfxRect *a, GfxRect *b);
    
    
void _font_draw(int fid, const char *txt, float x, float y) {
    FontInfo &fnt = _font->data[fid];
    
    
    
    glBindTexture(GL_TEXTURE_2D, gfx_tex_get_id(fnt.tex));
    
    int l = strlen(txt);
    int i;
    
    static float coords[256][12];
    static float verts[256][12];
    int c = 0;
    
    float tw = fnt.rect.w;
    float th = fnt.rect.h;
    float h = fnt.height;
//     fprintf(stderr,"%f %f %f\n",tw,th,h);
    float rh = h;
    
    
    for (i=0; i<l; i++) {
        int n = (unsigned char)txt[i];
        float w = fnt.chars[n].w;
        float rw = w;
        

        
        GfxRect s = GfxRect(fnt.chars[n].x/tw,fnt.chars[n].y/th,w/tw,h/th);
        GfxRect d = GfxRect(x,y,rw,rh);

        x += fnt.chars[n].w;

        if (!_gfx_clip(&d,&s)) { continue; }
        
        //fprintf(stderr,"ok: %f %f %f %f\n",d.x,d.y,d.w,d.h);
        
        coords[c][0] = s.x; coords[c][1] = s.y;
        coords[c][2] = s.x+s.w; coords[c][3] = s.y;
        coords[c][4] = s.x; coords[c][5] = s.y+s.h;
        coords[c][6] = s.x; coords[c][7] = s.y+s.h;
        coords[c][8] = s.x+s.w; coords[c][9] = s.y;
        coords[c][10] = s.x+s.w; coords[c][11] = s.y+s.h;

        verts[c][0] = d.x; verts[c][1] = d.y;
        verts[c][2] = d.x+d.w; verts[c][3] = d.y;
        verts[c][4] = d.x; verts[c][5] = d.y+d.h;
        verts[c][6] = d.x; verts[c][7] = d.y+d.h;
        verts[c][8] = d.x+d.w; verts[c][9] = d.y;
        verts[c][10] = d.x+d.w; verts[c][11] = d.y+d.h;
        
//         fprintf(stderr,"%d: %d,%d,%c\n",c,(int)x,(int)y,(char)n);
        c++;
    }

    
    glVertexPointer(2, GL_FLOAT, 0, verts);
    glTexCoordPointer(2, GL_FLOAT, 0, coords);
    glDrawArrays(GL_TRIANGLES, 0, c*6);
}
    

void _font_wrap_calc(int fid, const char *s, float ww, FontWrap &wrap) {
    FontInfo &fnt = _font->data[fid];
    
    wrap.total = 0; int ts=0,te=0,ls=0,lw=0;
    
    float w = 0;
    int len = strlen(s);
    for (int i=0; i<=len; i++) {
        te = i;
        int n = (unsigned char)s[i];
        float cw = fnt.chars[n].w;
        
        if (n != '\n' && n != 0) {
            if (ww == 0 || (w+cw) <= ww) {
                if (n == ' ') { ls = i; lw = w; }
                w += cw;
                continue;
            }
            // we can go back to a space
            if (ls) { te = ls; w = lw; n = s[te]; }
        }
        
        // else it's a linebreak ..
        wrap.lines[wrap.total][0] = ts;
        wrap.lines[wrap.total][1] = te;
        wrap.lines[wrap.total][2] = w;
        wrap.total ++;
        if (wrap.total >= FONT_LINES_MAX) { return; }
        
        // reset stuff
        ts = te + (n == ' ' || n == '\n' || n == 0); i = ts - 1; w = 0; ls = 0;
    }
}



float font_get_height(int fid, const char *_s, float w) {
    FontInfo &fnt = _font->data[fid];
    if (!_s || w==0) { return fnt.height; }
    const char *s = _font_filter(_s);
    FontWrap wrap;
    _font_wrap_calc(fid,s,w,wrap);
    return wrap.total * fnt.height;
}

float font_get_width(int fid, const char *_s) {
    FontInfo &fnt = _font->data[fid];
    const char *s = _font_filter(_s);
    
    float w = 0;
    int len = strlen(s);
    for (int i=0; i<len; i++) {
        int n = (unsigned char)s[i];
        w += fnt.chars[n].w;
    }
    return w;
}

void font_draw(int fid, const char *_s, float x, float y, float w, int align) {
    FontInfo &fnt = _font->data[fid];
    const char *s = _font_filter(_s);
    FontWrap wrap;
    _font_wrap_calc(fid,s,w,wrap);
    
    char * tmp = new char[strlen(s)+1]; // HACK: this avoids a crash on win32, so we have a bug somewhere

    for (int i=0; i<wrap.total; i++) {
        strcpy(tmp,s);
        tmp[wrap.lines[i][1]]=0;
        float xx = x;
        if (w) {
            float dx = w-wrap.lines[i][2];
            xx = x + dx*(align+1)/2;
        }
        GfxRect rect = GfxRect(xx,y,wrap.lines[i][2],fnt.height);
        // add in some smarts here, skip a line of text if none of it will
        // be rendered.  should speed up long list rendering.
        if (_gfx_clip(&rect,0)) {
            _font_draw(fid,&tmp[wrap.lines[i][0]],xx,y);
        } else {
            //fprintf(stderr,"skipped: %s\n",&tmp[wrap.lines[i][0]]);
        }
        y += fnt.height;
    }
    
    delete [] tmp;
}

GfxRect font_crect(int fid, const char *_s, int pos, float w, int align) {
    FontInfo &fnt = _font->data[fid];
    const char *s = _font_filter(_s);
    FontWrap wrap;
    _font_wrap_calc(fid,s,w,wrap);

    float x = 0, y = 0;
    float xx = 0;
    for (int i=0; i<wrap.total; i++) {
        if (w) {
            float dx = w-wrap.lines[i][2];
            xx = x + dx*(align+1)/2;
        }
        for (int n = wrap.lines[i][0]; n <= wrap.lines[i][1]; n++) {
            unsigned char c = s[n];
            if (n >= pos) {
                return GfxRect(xx,y,fnt.chars[c].w,fnt.height);
            }
            xx += fnt.chars[c].w;
        }
        y += fnt.height;
    }
    return GfxRect(xx,y-fnt.height,0,fnt.height);
}

int _font_dist(GfxRect &r,int x, int y) {
    int dx = (r.x)-x;
    int dy = (r.y+r.h/2)-y;
    return abs(dy)*65536 + abs(dx);
}

int font_cnear(int fid, const char *_s, int _x, int _y, float w, int align) {
    FontInfo &fnt = _font->data[fid];
    const char *s = _font_filter(_s);
    FontWrap wrap;
    _font_wrap_calc(fid,s,w,wrap);
    
    int bn = 0, bdist = -1;
    
    float x = 0, y = 0;
    float xx = 0;
    for (int i=0; i<wrap.total; i++) {
        if (w) {
            float dx = w-wrap.lines[i][2];
            xx = x + dx*(align+1)/2;
        }
        for (int n = wrap.lines[i][0]; n <= wrap.lines[i][1]; n++) {
            unsigned char c = s[n];
            GfxRect rect = GfxRect(xx,y,fnt.chars[c].w,fnt.height);
            int dist = _font_dist(rect,_x,_y);
            if (bdist == -1 || dist < bdist) {
                bdist = dist; bn = n;
            }
            xx += fnt.chars[c].w;
        }
        y += fnt.height;
    }
    return bn;
}
