#include "cuz_gfx.h"
#include "cuz_driver.h"
#include "cuz_opengl.h"

#include "cuz_json.h"
#include <map>
#include <math.h>
#ifndef M_PI
#define M_PI 3.1415926535897931
#endif

// private
#include "cuz_json.h"
struct _GfxData {
    GfxRect size;
    float orient,scale;
    GfxRect focus;
    GfxRect shift;
    
    var img_json; // the JSON file data
    std::map<std::string,int> img_map; // a map from a name to an id
    GfxImage img_data[2048]; // the C array using id's
    int img_n; // auto increment id
    
    var tex_json;
    std::map<std::string,int> tex_map;
    GfxTexture tex_data[256];
    int tex_n;
    
    GfxRect translate;
    GfxRect clip[8];
    int clip_n;
    
    _GfxData() {
        memset(&img_data,0,sizeof(img_data));
        img_n = 1;
        
        memset(&tex_data,0,sizeof(tex_data));
        tex_n = 1;
        
        memset(&clip,0,sizeof(clip));
        clip_n = 0;
    }
};

_GfxData *_gfx = 0;

void gfx_init() {
    _gfx = new _GfxData();

    glEnable(GL_TEXTURE_2D);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
//     glEnable(GL_DEPTH_TEST);
//     glDepthFunc(GL_LEQUAL);
//     glDepthFunc(GL_LESS);
    
    std::string json;
     
    json = driver_data_load("data.json");
    json_decode(_gfx->img_json,json.c_str(),json.size());

    json = driver_data_load("data.json");
    json_decode(_gfx->tex_json,json.c_str(),json.size());
    
    gfx_set_size(256,256);
    gfx_set_orient(0);
    gfx_set_scale(1.0);
    
//     _gfx->img_map = var(Json::objectValue);
//     _gfx->tex_map = var(Json::objectValue);
}

void gfx_deinit() {
    delete _gfx;
    _gfx = 0;
}

void gfx_clear() {
    glClearColor(0,0,0,0);
#ifndef PLATFORM_IOS
    glClearDepth(1.0f);
#endif
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void gfx_set_size(int w, int h) {
    _gfx->size.w = w;
    _gfx->size.h = h;
}

void gfx_get_size(int *w, int *h) {
    *w = _gfx->size.w;
    *h = _gfx->size.h;
}


void gfx_set_orient(float a) {
    _gfx->orient = a;
    
}

void gfx_set_scale(float s) {
    _gfx->scale = s;
}

float gfx_get_scale() {
    return _gfx->scale;
}

float gfx_get_orient() {
    return _gfx->orient;
}

GfxRect gfx_pos_to_hwpos(GfxRect p) {
    // data
    int rw,rh; driver_get_size(&rw,&rh);
    int vw,vh; gfx_get_size(&vw,&vh);
    float a = gfx_get_orient()*M_PI/180.0;
    float s = gfx_get_scale();
    
    // shift
//    p.x += _gfx->shift.x;
//    p.y += _gfx->shift.y;

    // center
    p.x = p.x - vw/2;
    p.y = p.y - vh/2;
    
    // unrotate
    float x = p.x, y = p.y;
    p.x = x*cos(-a) - y*sin(-a);
    p.y = x*sin(-a) + y*cos(-a);  

    // unscale
    p.x = p.x * s;
    p.y = p.y * s;
    
    // uncenter
    p.x = p.x + rw/2;
    p.y = p.y + rh/2;
    
    return p;
}

GfxRect gfx_hwpos_to_pos(GfxRect p) {
    // data
    int rw,rh; driver_get_size(&rw,&rh);
    int vw,vh; gfx_get_size(&vw,&vh);
    float a = gfx_get_orient()*M_PI/180.0;
    float s = gfx_get_scale();
    
    // center
    p.x = p.x - rw/2;
    p.y = p.y - rh/2;

    // scale
    p.x = p.x / s;
    p.y = p.y / s;

    // rotate
    float x = p.x, y = p.y;
    p.x = x*cos(a) - y*sin(a);
    p.y = x*sin(a) + y*cos(a);  

    // uncenter
    p.x = p.x + vw/2;
    p.y = p.y + vh/2;

    // unshift
//    p.x -= _gfx->shift.x;
//    p.y -= _gfx->shift.y;
    
    return p;
}

GfxRect gfx_hwrect_to_rect(GfxRect r) {
    GfxRect p1 = gfx_hwpos_to_pos(GfxRect(r.x,r.y,0,0));
    GfxRect p2 = gfx_hwpos_to_pos(GfxRect(r.x+r.w,r.y+r.h,0,0));
    return GfxRect(p1.x,p1.y,p2.x-p1.x,p2.y-p1.y);
}


GfxRect gfx_rect_to_hwrect(GfxRect r) {
    GfxRect p1 = gfx_pos_to_hwpos(GfxRect(r.x,r.y,0,0));
    GfxRect p2 = gfx_pos_to_hwpos(GfxRect(r.x+r.w,r.y+r.h,0,0));
    return GfxRect(p1.x,p1.y,p2.x-p1.x,p2.y-p1.y);
}


void gfx_event_correct(Event *e) {
    if (e->type == EVT_MOTION || e->type == EVT_DOWN || e->type == EVT_UP) {
        e->_x = e->x;
        e->_y = e->y;
        e->_dx = e->dx;
        e->_dy = e->dy;
        
        GfxRect r = gfx_hwrect_to_rect(GfxRect(e->_x,e->_y,e->_dx,e->_dy));
        e->x = r.x; e->y = r.y;
        e->dx = r.w; e->dy = r.h;
        
        //fprintf(stderr,"%d %d %d %d \n",(int)e->x,(int)e->y,(int)e->dx,(int)e->dy);
    }
}


void gfx_focus(GfxRect r) {
    _gfx->focus = r;
}

void gfx_blur() {
    _gfx->focus = GfxRect(0,0,0,0);
}

void gfx_ortho() {
    // data
    int rw,rh; driver_get_size(&rw,&rh);
    int vw,vh; gfx_get_size(&vw,&vh);
    float a = gfx_get_orient()*M_PI/180.0;
    float s = gfx_get_scale();

    // identity
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0,rw,rh,0);
    
    // center
    glTranslatef(rw/2,rh/2,0);
    
    // scale
    glScalef(s,s,1);

    // rotate
    glRotatef(-a*180/M_PI,0,0,1);
    
    // uncenter
    glTranslatef(-vw/2,-vh/2,0);
    
    // shift
//    _gfx->shift.y = 0;
//    if (_gfx->focus.h) {
//        int ky = vh - driver_keyboard_get_height() * s;
//        int fy = _gfx->focus.y + _gfx->focus.h;
//        if (fy > ky) {
//            _gfx->shift.y = ky-fy;
//        }
//    }
//    glTranslatef(_gfx->shift.x,_gfx->shift.y,0);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

inline bool _gfx_clip_test(GfxRect &clip, GfxRect *rect, GfxRect *coords) {
//    fprintf(stderr,"clip: %f %f %f %f, rect: %f %f %f %f\n",
//            clip.x,clip.y,clip.w,clip.h, rect->x,rect->y,rect->w,rect->h);
    
    if (rect->w <= 0 || rect->h <= 0) { return 0; }
    
    if (rect->x < clip.x) {
        float dx = clip.x - rect->x;
        if (coords) {
            coords->x += dx * coords->w / rect->w;
            coords->w -= dx * coords->w / rect->w;
        }
        rect->x += dx;
        rect->w -= dx;
    }
    if (rect->w <= 0) { return 0; }
    
    if ((rect->x+rect->w) > (clip.x + clip.w)) {
        float dx = rect->x+rect->w - (clip.x+clip.w);
        if (coords) {
            coords->w -= dx * coords->w / rect->w;
        }
        rect->w -= dx;
    }
    if (rect->w <= 0) { return 0; }
 
    if (rect->y < clip.y) {
        float dy = clip.y - rect->y;
        if (coords) {
            coords->y += dy * coords->h / rect->h;
            coords->h -= dy * coords->h / rect->h;
        }
        rect->y += dy;
        rect->h -= dy;
    }
    if (rect->h <= 0) { return 0; }

    if ((rect->y+rect->h) > (clip.y + clip.h)) {
        float dy = rect->y+rect->h - (clip.y+clip.h);
        if (coords) {
            coords->h -= dy * coords->h / rect->h;
        }
        rect->h -= dy;
    }
    if (rect->h <= 0) { return 0; }
    
    return 1;
}

void gfx_translate(int x, int y) {
    _gfx->translate.x += x;
    _gfx->translate.y += y;
    glTranslatef(x,y,0);
}

void gfx_clip(int x, int y, int w, int h) {
    int &n = _gfx->clip_n;
    n += 1;
    if (n > 8) { return; }
    
    _gfx->clip[n-1] = GfxRect(x+_gfx->translate.x,y+_gfx->translate.y,w,h);
    if (n >= 2) {
        _gfx_clip_test(_gfx->clip[n-2],&_gfx->clip[n-1],0);
    }
}

bool _gfx_clip(GfxRect *rect, GfxRect *coords) {
    int &n = _gfx->clip_n;
    if (!n) { return 1; }
    if (n > 8) { return 0; }
    GfxRect clip = _gfx->clip[n-1];
    clip.x -= _gfx->translate.x;
    clip.y -= _gfx->translate.y;
//    fprintf(stderr,"clip: %f %f %f %f\n",clip.x,clip.y,clip.w,clip.h); 
//    fprintf(stderr,"rect: %f %f %f %f\n",rect->x,rect->y,rect->w,rect->h);
//    fprintf(stderr,"cord: %f %f %f %f\n",coords->x,coords->y,coords->w,coords->h);
    return _gfx_clip_test(clip,rect,coords);
}

void gfx_unclip() {
    int &n = _gfx->clip_n;
    if (!n) { return; }
    n -= 1;    
}


void gfx_tex_blit(int tid, int dx, int dy, int dw, int dh) {
    GfxRect s = GfxRect(0,0,1,1);
    GfxRect d = GfxRect(dx,dy,dw,dh);
    if (!_gfx_clip(&d,&s)) { return; }
    
    float coords[] ={s.x,s.y, s.x+s.w,s.y, s.x,s.y+s.h, s.x+s.w,s.y+s.h};
    float verts[] = {d.x,d.y, d.x+d.w,d.y, d.x,d.y+d.h, d.x+d.w,d.y+d.h};

    glBindTexture(GL_TEXTURE_2D, tid);
    glVertexPointer(2, GL_FLOAT, 0, verts);
    glTexCoordPointer(2, GL_FLOAT, 0, coords);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
}

void gfx_tex_blit2(int tid, float sx, float sy, float sw, float sh, float dx, float dy, float dw, float dh) {
    GfxRect s = GfxRect(sx,sy,sw,sh);
    GfxRect d = GfxRect(dx,dy,dw,dh);
    if (!_gfx_clip(&d,&s)) { return; }
    
    float coords[] ={s.x,s.y, s.x+s.w,s.y, s.x,s.y+s.h, s.x+s.w,s.y+s.h};
    float verts[] = {d.x,d.y, d.x+d.w,d.y, d.x,d.y+d.h, d.x+d.w,d.y+d.h};
    
    glBindTexture(GL_TEXTURE_2D, tid);
    glVertexPointer(2, GL_FLOAT, 0, verts);
    glTexCoordPointer(2, GL_FLOAT, 0, coords);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
}    

/*
   Quads:
    3  4
    1  2
*/
void gfx_tex_quad(int tid,
    float x1,float y1, float z1,
    float x2,float y2, float z2,
    float x3,float y3, float z3,
    float x4,float y4, float z4
) {
    float coords[] ={ 0,1, 1,1, 0,0, 1,0 };
    float verts[] = { x1,y1,z1, x2,y2,z2, x3,y3,z3, x4,y4,z4 };
    
    glBindTexture(GL_TEXTURE_2D, tid);
    glVertexPointer(3, GL_FLOAT, 0, verts);
    glTexCoordPointer(2, GL_FLOAT, 0, coords);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
}

GfxRect _gfx_json_rect(var &js) {
    GfxRect r;
    r.x = js["x"];
    r.y = js["y"];
    r.w = js["w"];
    r.h = js["h"];
    return r;
}

int gfx_tex_load(const char *name) {
    var &js = _gfx->tex_json;

    if (_gfx->tex_map.find(name) == _gfx->tex_map.end()) {
        fprintf(stderr,"trying load tex %s\n",name);
        if (!js.has(name)) {
            fprintf(stderr,"unable to find tex %s\n",name);
            _gfx->tex_map[name] = 0;
            return 0;
        }
        int n = _gfx->tex_n;
        fprintf(stderr,"doing load tex %s (%d)\n",name,n);
        _gfx->tex_map[name] = n;
        _gfx->tex_data[n].n = n;
        _gfx->tex_data[n].tex = driver_tex_load(js[name]["fname"].c_str());
        _gfx->tex_data[n].rect = _gfx_json_rect(js[name]["rect"]);
        _gfx->tex_n += 1;
    }
    
    return _gfx->tex_map[name];
}




int gfx_img_load(const char *name) {
    var &js = _gfx->img_json;

    if (_gfx->img_map.find(name) == _gfx->img_map.end()) {
        fprintf(stderr,"trying load img %s\n",name);
        if (!js.has(name)) {
            fprintf(stderr,"unable to find img %s\n",name);
            _gfx->img_map[name] = 0;
            return 0;
        }
        int n = _gfx->img_n;
        fprintf(stderr,"doing load img %s (%d)\n",name,n);
		fprintf(stderr,"tex: %s\n",js[name]["texture"].c_str());
        _gfx->img_map[name] = n;
        _gfx->img_data[n].n = n;
        _gfx->img_data[n].tid = gfx_tex_load(js[name]["texture"].c_str());
        _gfx->img_data[n].rect = _gfx_json_rect(js[name]["rect"]);
        _gfx->img_data[n].shape = _gfx_json_rect(js[name]["shape"]);
        _gfx->img_n += 1;
    }
    
    return _gfx->img_map[name];
}

void gfx_img_blit(int _img, float dx, float dy, float dw, float dh) {
    GfxImage &img = _gfx->img_data[_img];
    GfxTexture &tex = _gfx->tex_data[img.tid];
    if (!dw) { dw = img.shape.w; }
    if (!dh) { dh = img.shape.h; }
    
    GfxRect s = GfxRect(img.rect.x / tex.rect.w, img.rect.y / tex.rect.h, img.rect.w / tex.rect.w, img.rect.h / tex.rect.h);
    GfxRect d = GfxRect(dx,dy,dw,dh);
    
    if (!_gfx_clip(&d,&s)) { return; }
    
    float coords[] ={s.x,s.y, s.x+s.w,s.y, s.x,s.y+s.h, s.x+s.w,s.y+s.h};
    float verts[] = {d.x,d.y, d.x+d.w,d.y, d.x,d.y+d.h, d.x+d.w,d.y+d.h};
    glBindTexture(GL_TEXTURE_2D, tex.tex);
    glVertexPointer(2, GL_FLOAT, 0, verts);
    glTexCoordPointer(2, GL_FLOAT, 0, coords);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
}

void gfx_img_quad(int _img,
    float x1,float y1, float z1,
    float x2,float y2, float z2,
    float x3,float y3, float z3,
    float x4,float y4, float z4
) {
    GfxImage &img = _gfx->img_data[_img];
    GfxTexture &tex = _gfx->tex_data[img.tid];
    
    float cx1 = img.rect.x / tex.rect.w;
    float cy1 = img.rect.y / tex.rect.h;
    float cx2 = (img.rect.x + img.rect.w) / tex.rect.w;
    float cy2 = (img.rect.y + img.rect.h) / tex.rect.h;
    
    float coords[] ={ cx1,cy2, cx2,cy2, cx1,cy1, cx2,cy1 };
    
    float verts[] = { x1,y1,z1, x2,y2,z2, x3,y3,z3, x4,y4,z4 };
    glBindTexture(GL_TEXTURE_2D, tex.tex);
    glVertexPointer(3, GL_FLOAT, 0, verts);
    glTexCoordPointer(2, GL_FLOAT, 0, coords);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
}

int gfx_img_get_tex(int _img) { 
    GfxImage &img = _gfx->img_data[_img];
    GfxTexture &tex = _gfx->tex_data[img.tid];
    return tex.tex;
}
int gfx_img_get_width(int _img) { 
    GfxImage &img = _gfx->img_data[_img];
    return img.shape.w;
}
int gfx_img_get_height(int _img) { 
    GfxImage &img = _gfx->img_data[_img];
    return img.shape.h;
}
int gfx_tex_get_id(int tid) { 
    GfxTexture &tex = _gfx->tex_data[tid];
    return tex.tex;
}


void gfx_fill(unsigned int c, int dx, int dy, int dw, int dh) {
    GfxRect d = GfxRect(dx,dy,dw,dh);
    if (!_gfx_clip(&d,0)) { return; }
/*    fprintf(stderr,"%f %f %f %f %x\n",d.x,d.y,d.w,d.h,c);*/
    float verts[] = {d.x,d.y, d.x+d.w,d.y, d.x,d.y+d.h, d.x+d.w,d.y+d.h};
    glColor4ub(((c>>16)&0xff),((c>>8)&0xff),((c>>0)&0xff),255);
    glDisable(GL_TEXTURE_2D);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    
    glVertexPointer(2, GL_FLOAT, 0, verts);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    
    glEnable(GL_TEXTURE_2D);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    glColor4ub(255,255,255,255);
}

void gfx_tex_box(int tid, int rx, int ry, int rw, int rh, int sz) {
    for (float yy=ry; yy<ry+rh; yy+=sz) {
        float sy = 0; 
        float dy = yy;
        float sh = sz;
        float dh = sz;
        if (yy>ry) { sy += sz; }
        if (yy>=(ry+rh-sz*2)&&yy<(ry+rh-sz)) {
            sh = (ry+rh-sz)-yy;
            dh = sh;
        }
        if (yy>=(ry+rh-sz)) {
            sy += sz;
            dy = ry+rh-sz;
        }
        
        for (float xx=rx; xx<rx+rw; xx+=sz) {
            float sx = 0;
            float dx = xx;
            float sw = sz;
            float dw = sz;
            if (xx > rx) { sx += sz; }
            if (xx>=(rx+rw-sz*2)&&xx<(rx+rw-sz)) {
                sw = (rx+rw-sz)-xx;
                dw = sw;
            }
            if (xx>=(rx+rw-sz)) {
                sx += sz;
                dx = rx+rw-sz;
            }
            
            gfx_tex_blit2(tid,sx/(sz*3),sy/(sz*3),sw/(sz*3),sh/(sz*3),dx,dy,dw,dh);
        }
    }
    
}
    
