#ifndef CUZ_GFX_H
#define CUZ_GFX_H

#include "cuz_event.h"

struct GfxRect {
    float x,y,w,h;
    GfxRect() : x(0),y(0),w(0),h(0) { }
    GfxRect(float _x, float _y, float _w, float _h) : x(_x),y(_y),w(_w),h(_h) { }
    
    void normalize() {
        if (w<0) { x+=w; w=-w; }
        if (h<0) { y+=h; h=-h; }
    }
};

struct GfxImage {
    int n;
    int tid;
    GfxRect rect;
    GfxRect shape;
};

struct GfxTexture {
    int n;
    int tex;
    GfxRect rect;
};

void gfx_init();
void gfx_deinit();
void gfx_clear();

//void gfx_ortho(int w, int h);

GfxRect gfx_rect_to_hwrect(GfxRect r);
void gfx_set_size(int w, int h);
void gfx_set_orient(float a);
void gfx_set_scale(float s);

void gfx_event_correct(Event *e);
void gfx_ortho();

// when the keyboard pops up / down, we can move the screen around based on a focus region
void gfx_focus(GfxRect r); 
void gfx_blur();

int gfx_tex_load(const char *fname);
int gfx_tex_get_id(int tid);

void gfx_translate(int x, int y);
void gfx_clip(int x, int y, int w, int h);
bool _gfx_clip(GfxRect *rect, GfxRect *coords) ;
void gfx_unclip();

void gfx_tex_blit(int tid, int x, int y, int w, int h);
// void gfx_tex_blit(int tid, float x, float y, float w, float h);
void gfx_tex_blit2(int tid, float sx, float sy, float sw, float sh, float dx, float dy, float dw, float dh);
void gfx_tex_quad(int tid,
    float x1,float y1, float z1,
    float x2,float y2, float z2,
    float x3,float y3, float z3,
    float x4,float y4, float z4
);

int gfx_img_load(const char *name);
int gfx_img_get_width(int img);
int gfx_img_get_height(int img);
int gfx_img_get_tex(int img);
// void gfx_img_blit(int img, int x, int y, int w=0, int h=0);
void gfx_img_blit(int img, float x, float y, float w=0, float h=0);
void gfx_img_quad(int img,
    float x1,float y1, float z1,
    float x2,float y2, float z2,
    float x3,float y3, float z3,
    float x4,float y4, float z4
);
void gfx_fill(unsigned int c, int x, int y, int w, int h);
void gfx_tex_box(int img, int x, int y, int w, int h, int sz);


#endif
