#include "cuz_gui.h"
#include "cuz_opengl.h"

void Theme::resize(Widget *w) {
    StyleSheet &style = w->style;
    STYLE_GET_SIZES;
    
    w->resize();
    
    w->rect.x = left;
    w->rect.y = top;
    w->_rect.w = left+w->rect.w+right;
    w->_rect.h = top+w->rect.h+bottom;
}

void Theme::resize(Widget *w, int width, int height) {
    StyleSheet &style = w->style;
    STYLE_GET_SIZES;
    
    w->resize(width-(left+right),height-(top+bottom));
    
    w->rect.x = left;
    w->rect.y = top;
    w->_rect.w = left+w->rect.w+right;
    w->_rect.h = top+w->rect.h+bottom;
}


int Theme::event(Widget *w, Event *e_) {
    StyleSheet &style = w->style;
    
    Event e = *e_;
    e.x -= w->rect.x;
    e.y -= w->rect.y;
    
    int snd = GUI_NONE;
    if (e_->type == EVT_CLICK) { snd = STYLE(sfx_onclick,GUI_NONE); }
    if (e_->type == EVT_DOWN) { snd = STYLE(sfx_ondown,GUI_NONE); }
    if (e_->type == EVT_UP) { snd = STYLE(sfx_onup,GUI_NONE); }
//     if (snd != GUI_NONE) { Game::game->sound_play(snd); }
    
    return w->event(&e);
}

void Theme::paint(Widget *w) {
    StyleSheet &style = w->style;
    MyRect &rect = w->_rect;

    int a;
    int c = STYLE(background,GUI_NONE);
    int top = STYLE(margin[STYLE_TOP],0);
    int right = STYLE(margin[STYLE_RIGHT],0);
    int bottom = STYLE(margin[STYLE_BOTTOM],0);
    int left = STYLE(margin[STYLE_LEFT],0);
    
    int topb = STYLE(border[STYLE_TOP],0);
    int rightb = STYLE(border[STYLE_RIGHT],0);
    int bottomb = STYLE(border[STYLE_BOTTOM],0);
    int leftb = STYLE(border[STYLE_LEFT],0);
    
    if (c != GUI_NONE) {
        float dx=left+leftb,dy=top+topb,dw=rect.w-(left+right+leftb+rightb),dh=rect.h-(top+bottom+topb+bottomb);
        gfx_fill(c,dx,dy,dw,dh);
    }
    c = STYLE(border_color,GUI_NONE);
    if (c != GUI_NONE) {
        float dx,dy,dw,dh;
        if ((dh=topb)!=0) {
            dx=left;dy=top;dw=rect.w-(left+right);
            gfx_fill(c,dx,dy,dw,dh);
        }
        if ((dh=bottomb)!=0) {
            dx=left;dy=rect.h-(dh+bottom);dw=rect.w-(left+right);
            gfx_fill(c,dx,dy,dw,dh);
        }
        if ((dw=rightb)!=0) {
            dx=rect.w-(dw+right);dy=top+topb;dh=rect.h-(top+bottom+topb+bottomb);
            gfx_fill(c,dx,dy,dw,dh);
        }
        if ((dw=leftb)!=0) {
            dx=left;dy=top+topb;dh=rect.h-(top+bottom+topb+bottomb);
            gfx_fill(c,dx,dy,dw,dh);
        }
    }
    
    c = STYLE(box_color,GUI_NONE);
    int tex=STYLE(box_tex,GUI_NONE);
    if (c!=GUI_NONE && tex != GUI_NONE) {
        float grow = STYLE(box_grow,0);
        float size = STYLE(box_size,16);
        float dx=left,dy=top,dw=rect.w-(left+right),dh=rect.h-(top+bottom);
        a = 255;
        glColor4ub(((c>>16)&0xff),((c>>8)&0xff),((c>>0)&0xff),a);
        gfx_tex_box(gfx_img_get_tex(tex),dx-grow,dy-grow,dw+grow*2,dh+grow*2,(int)size);
        glColor4ub(255,255,255,255);
    }
    
    int img = STYLE(img,GUI_NONE);
    if (img != GUI_NONE) {
        c = STYLE(img_color,0xffffff);
        a = STYLE(img_alpha,0xff);
         glColor4ub(((c>>16)&0xff),((c>>8)&0xff),((c>>0)&0xff),a);
         //int af = STYLE(img_autofit,1);
        int af = 1;
         int ww = af?w->rect.w:0,hh=af?w->rect.h:0;
        gfx_img_blit(img,w->rect.x,w->rect.y,ww,hh);
         glColor4ub(255,255,255,255);
    }
    
    int ovr = STYLE(ovr,GUI_NONE);
    if (ovr != GUI_NONE) {
        float grow=0;
        float size = STYLE(ovr_size,16);
        c = STYLE(ovr_color,0xffffff);
        a = STYLE(ovr_alpha,0xff);
        glColor4ub(((c>>16)&0xff),((c>>8)&0xff),((c>>0)&0xff),a);
        float dx=left,dy=top,dw=rect.w-(left+right),dh=rect.h-(top+bottom);
        gfx_tex_box(gfx_img_get_tex(ovr),dx-grow,dy-grow,dw+grow*2,dh+grow*2,(int)size);
        //        gfx_img_blit(ovr,w->rect.x,w->rect.y,w->rect.w,w->rect.h);
        glColor4ub(255,255,255,255);
    }
    
    // we'll try and conserve stack, in the case of a container with an overlay, this will not work out so well
//     glPushMatrix();
    //glTranslatef(w->rect.x,w->rect.y,0);
    gfx_translate(w->rect.x,w->rect.y);
    w->paint();
    gfx_translate(-w->rect.x,-w->rect.y);
    //glTranslatef(-w->rect.x,-w->rect.y,0);
//     glPopMatrix();
    

}

