#ifndef CUZ_GUI_H
#define CUZ_GUI_H

#include <vector>
#include <algorithm>
#include <string>

#include <math.h>

#ifndef M_PI
#define M_PI 3.1415926535897931
#endif

#include "cuz_util.h"
struct MyRect {
    int x,y,w,h;
    MyRect(int x, int y, int w, int h) {
        this->x=x; this->y=y;
        this->w=w; this->h=h;
    }
    MyRect() { MyRect(0,0,0,0); }
};
#include "cuz_event.h"

#define GUI_NONE -0xffffff

struct Widget;


enum { STYLE_TOP,STYLE_RIGHT,STYLE_BOTTOM,STYLE_LEFT};
#define STYLE_CALC(v) STYLE(padding[v],0)+STYLE(border[v],0)+STYLE(margin[v],0)
#define STYLE_GET_SIZES \
    int top = STYLE_CALC(STYLE_TOP); \
    int right = STYLE_CALC(STYLE_RIGHT); \
    int bottom = STYLE_CALC(STYLE_BOTTOM); \
    int left = STYLE_CALC(STYLE_LEFT);

#define STYLE_SET(t) \
    void set_##t(int v, int v2=GUI_NONE, int v3=GUI_NONE, int v4=GUI_NONE) { \
        t[0]=v; t[1]=(v2!=GUI_NONE)?v2:v; t[2]=(v3!=GUI_NONE)?v3:v; t[3]=(v4!=GUI_NONE)?v4:v; \
    }

struct StyleData {
    // misc properties
    int width,height; int clip;
    int align,valign;
    
    // color properties
    int color,background;
    
    // box properties
    int border[4], border_color, padding[4], margin[4];
    
    // my style props
    int box_color; int box_grow,box_size,box_tex;
    
    // img data
    int img; int img_color; int img_alpha; int img_blend; int img_autofit;
    
    // overlay data (images)
    int ovr; int ovr_color; int ovr_alpha; int ovr_blend; int ovr_size;
    
    // sfx
    int sfx_onclick; int sfx_ondown; int sfx_onup;
    
    StyleData() :
        width(GUI_NONE),height(GUI_NONE),clip(GUI_NONE),
        align(GUI_NONE),valign(GUI_NONE),
        color(GUI_NONE),background(GUI_NONE),
        border_color(GUI_NONE),box_color(GUI_NONE), box_grow(GUI_NONE), box_size(GUI_NONE),
        box_tex(GUI_NONE),
        img(GUI_NONE), img_color(GUI_NONE), img_alpha(GUI_NONE),
        img_blend(GUI_NONE), img_autofit(GUI_NONE),
        ovr(GUI_NONE), ovr_color(GUI_NONE), ovr_alpha(GUI_NONE), ovr_blend(GUI_NONE),
        ovr_size(GUI_NONE),
        sfx_onclick(GUI_NONE), sfx_ondown(GUI_NONE),sfx_onup(GUI_NONE)
        {
        set_margin(GUI_NONE);
        set_padding(GUI_NONE);
        set_border(GUI_NONE);
        
//         set_border(1); border_color=0xff0000;
    }
    
    STYLE_SET(padding);
    STYLE_SET(margin);
    STYLE_SET(border);
    

};

enum { STYLE_NONE,STYLE_CUR,STYLE_HOVER,STYLE_FOCUS,STYLE_ACTIVE,STYLE_ALT,STYLE_MAX };

#define _STYLE_TRANS(k,fnc,c,v) \
    if (v != GUI_NONE) { \
        if (data[STYLE_CUR].k == GUI_NONE) { \
            data[STYLE_CUR].k = v; \
        } else { \
            data[STYLE_CUR].k = fnc(data[STYLE_CUR].k,v,c); \
        } \
    }

#define _STYLE_PREP(k,fnc,c) \
    if (alt && data[STYLE_ALT].k != GUI_NONE) { \
        _STYLE_TRANS(k,fnc,c,data[STYLE_ALT].k); \
    } else if (data[state].k != GUI_NONE) { \
        _STYLE_TRANS(k,fnc,c,data[state].k); \
    } else { \
        _STYLE_TRANS(k,fnc,c,this->k); \
    }

struct StyleSheet : StyleData {
    StyleData data[STYLE_MAX];
    int hover,focus,active,alt;
    int state;
    StyleSheet() : StyleData(),
        hover(0),focus(0),active(0),alt(0),state(STYLE_NONE)
        
        {
        // set some sane defaults for core style ..
        //border_color=0;background=0;box_color=0;
        }
    void restate() {
        active = (focus && hover);
//         if (alt) { state = STYLE_ALT; return; }
        if (active) { state = STYLE_ACTIVE; return; }
        if (focus) { state = STYLE_FOCUS; return; }
        if (hover) { state = STYLE_HOVER; return; }
        state = STYLE_NONE;
        return;
    }
    void loop() {
        _STYLE_PREP(width,i_prep,1000);
        _STYLE_PREP(height,i_prep,1000);
        _STYLE_PREP(clip,n_prep,0);
        _STYLE_PREP(align,n_prep,0);
        _STYLE_PREP(valign,n_prep,0);
        _STYLE_PREP(color,c_prep,36);
        _STYLE_PREP(background,c_prep,36);
        for(int i=0;i<4;i++){
            _STYLE_PREP(margin[i],i_prep,1);
            _STYLE_PREP(border[i],i_prep,1);
            _STYLE_PREP(padding[i],i_prep,1);
        }
        _STYLE_PREP(border_color,c_prep,36);
        _STYLE_PREP(box_color,c_prep,36);
        _STYLE_PREP(box_grow,i_prep,1);
        _STYLE_PREP(box_size,i_prep,1);
        _STYLE_PREP(box_tex,n_prep,0);
        _STYLE_PREP(img,n_prep,0);
        _STYLE_PREP(img_color,c_prep,36);
        _STYLE_PREP(img_alpha,i_prep,36);
        _STYLE_PREP(img_blend,n_prep,0);
        _STYLE_PREP(img_autofit,n_prep,0);
        _STYLE_PREP(ovr,n_prep,0);
        _STYLE_PREP(ovr_color,c_prep,36);
        _STYLE_PREP(ovr_size,i_prep,1);
        _STYLE_PREP(ovr_alpha,i_prep,36);
        _STYLE_PREP(ovr_blend,n_prep,0);
        _STYLE_PREP(sfx_onclick,n_prep,0);
        _STYLE_PREP(sfx_ondown,n_prep,0);
        _STYLE_PREP(sfx_onup,n_prep,0);
    }
    
    int i_prep(int a, int b, int c) {
        if (myabs(b-a) < c) { return b; }
        return a + mysign(b-a)*c;
    }
    
    int n_prep(int a, int b, int c) {
        return b;
    }
    
    int c_prep(int a, int b, int c) {
        return i_prep(a&0xff0000,b&0xff0000,c<<16) |
            i_prep(a&0xff00,b&0xff00,c<<8) |
            i_prep(a&0xff,b&0xff,c<<0);
    }
    
    
    
        
};

#define WSTYLE(w,k,d) (w->style.data[STYLE_CUR].k!=GUI_NONE?w->style.data[STYLE_CUR].k:d)
#define STYLE(k,d) (style.data[STYLE_CUR].k!=GUI_NONE?style.data[STYLE_CUR].k:d)

struct Widget;
struct Theme {
    static void resize(Widget *w);
    static void resize(Widget *w, int width, int height);
    static int event(Widget *w, Event *e);
    static void paint(Widget *w);
};



// #include "myrect.h"
#include "cuz_font.h"
#include "cuz_gfx.h"
#include "cuz_opengl.h"
#include "cuz_driver.h"


// #include "theme.h"

struct Font {
    int fid;
    Font() { fid = 0; }
    Font(int _fid) { fid = _fid; }
    int get_height(const char *s=0,int w = 0) { return font_get_height(fid,s,w); }
    int get_width(const char *s, int w=0) { return w?w:font_get_width(fid,s); }
    void draw(const char *s, int x, int y, int w = 0, int align=-1) { font_draw(fid,s,x,y,w,align); }
    GfxRect crect(const char *s, int pos, int w = 0, int align=-1) { 
        return font_crect(fid,s,pos,w,align);
    }
    int cnear(const char *s, int x, int y, int w = 0, int align=-1) {
        return font_cnear(fid,s,x,y,w,align);
    }
    

};


struct Container;

struct Widget {
    int _type;
    Container *parent;
    Widget *next;
    MyRect rect,_rect; MyRect view;
    int onclick; int ondown; int onmove; int onchange;
    std::string name;
    
//     static Game *game;
    
    StyleSheet style;
    int ignore;

    int b_col, b_row, b_colspan,b_rowspan, b_x, b_y;
//     int s_width,s_height;

//     const char *name;
    
    Widget() :
        _type(0),
        parent(0),next(0),
        rect(0,0,0,0),_rect(0,0,0,0),view(0,0,0,0),
        onclick(0),ondown(0),onmove(0),onchange(0),
        style(),ignore(0),
        b_col(0),b_row(0),b_colspan(1),b_rowspan(1)
//         name(0)
        {
    }
    
    virtual Widget *find(const char *v) {
        if (this->name == v) { return this; }
        return 0;
    }
    
    bool replace(const char *v, Widget *w);
    

    
    virtual const char *get_value_s() {
        fprintf(stderr,"const char *get_value_s(): n/a\n");
        return 0;
    }
    virtual void set_value_s(const char *) {
        fprintf(stderr,"set_value_s(const char *): n/a\n");
    }
    virtual int get_value_i() {
        fprintf(stderr,"int get_value_i(): n/a\n");
        return 0;
    }
    virtual void set_value_i(int) {
        fprintf(stderr,"set_value_i(int): n/a\n");
    }
    
    virtual inline void _focus() ;

    virtual int event(Event *e) {
//         fprintf(stderr,"widget::event\n");
        if (e->type == EVT_KEYDOWN && e->v == KEY_TAB && this->next != 0) {
            this->next->_focus();
        }

        if (e->type == EVT_CLICK) { return onclick; }
        if (e->type == EVT_MOTION) { return onmove; }
        if (e->type == EVT_DOWN) { return ondown; }
        if (e->type == EVT_FOCUS) {
            this->style.focus = 1; this->style.restate();
            return 0;
        }
        if (e->type == EVT_ENTER) {
            this->style.hover = 1; this->style.restate();
            return 0;
        }
        if (e->type == EVT_EXIT) {
            this->style.hover = 0; this->style.restate();
            return 0;
        }
        if (e->type == EVT_BLUR) {
            this->style.focus = 0; this->style.restate();
            return 0;
        }
        
        return 0;
    }
    
    virtual void loop() {
    }
    
    virtual void paint() {
//         fprintf(stderr,"widget::paint\n");
    }
    
    virtual void resize() {
        this->rect.h = STYLE(height,0);
        this->rect.w = STYLE(width,0);
    }
    virtual void resize(int width, int height) { this->resize(); }
    
    virtual int _event(Event *e) {
        return Theme::event(this,e);
    }
    virtual void _paint() {
        Theme::paint(this);
    }
    virtual void _resize() {
        Theme::resize(this);
    }
    virtual void _resize(int width, int height) {
        Theme::resize(this,width,height);
    }
    
    
    MyRect get_abs_rect(bool inc_style=false);
    
//    int _clip[5];
    void _enable_clip() {
        //MyRect r = get_abs_rect();
        //gfx_clip(r.x,r.y,r.w,r.h);
        gfx_clip(0,0,rect.w,rect.h);
    }
    
    void _disable_clip() {
        gfx_unclip();
    }
    
    virtual ~Widget() {  }
};


struct Container : Widget {
    std::vector<Widget *> widgets;
    Widget *hover;
    Widget *focus;
    
    Container() : Widget(),
        widgets(0),hover(0),focus(0)
    {
//         _type=1; // find/name
    }
    
/*    Widget *_find(const char *n) {
        unsigned int i;
        for (i=0; i<widgets.size(); i++) {
            Widget *w = widgets[i];
//             fprintf(stderr,"%x: %d = %x, name = %s\n",this,i,w,w->name);
            if (w->name != 0 && strcmp(w->name,n) == 0) { return w; }
            if (!w->_type) { continue; }
            Container *c = (Container *)w;
            w = c->_find(n);
            if (w) { return w; }
        }
        return 0;
    }
    
    Widget *find(const char *n) {
        Widget *w = this->_find(n);
        if (!w) {
            fprintf(stderr,"find(%s): failed\n",n);
        }
        return w;
    }*/
    
    
    void add(Widget *w, int x, int y) {
        w->_rect.x = x;
        w->_rect.y = y;
        widgets.push_back(w);
        w->parent = this;
        w->style.loop();
        w->resize();
    }
    
    
    void remove(Widget *w1) {
        if (!w1) {
//             fprintf(stderr,"w1 is null\n");
            return;
        }
        if (hover == w1) { hover = 0; }
        if (focus == w1) { focus = 0; }
        Widget *w;
        unsigned int i;
        for (i=0; i<widgets.size(); i++) {
            w = widgets[i];
            if (w == w1) {
//                 fprintf(stderr,"removed w1 %x\n",w1);
                widgets.erase(widgets.begin()+i);
                delete w1;
                return;
            }
        }
//         fprintf(stderr,"could not find widget %x\n",w1);
    }
    
    virtual Widget *find(const char *v) {
        for (unsigned int i=0; i<widgets.size(); i++) {
            Widget *w = widgets[i];
            Widget *r = w->find(v);
            if (r) { return r; }
        }
        return Widget::find(v);
    }
        

    virtual void resize() {
        Widget *w;
        unsigned int i,ww=0,hh=0;
        for (i=0; i<widgets.size(); i++) {
            w = widgets[i];
            w->style.loop();
            w->_resize();
            ww = (int)mymax(ww,w->_rect.x+w->_rect.w);
            hh = (int)mymax(hh,w->_rect.y+w->_rect.h);
        }
        this->rect.w = ww;
        this->rect.h = hh;
    }
        
    
    virtual void paint() {
        Widget *w;
        unsigned int i;
        unsigned int sz = widgets.size();
        int mtx = (sz>1 || parent==0); // if we're just doing one, don't bother, if we're top level, gotta save and restore
        
		//int depth;
		//glGetIntegerv(GL_MODELVIEW_STACK_DEPTH,&depth);
		//fprintf(stderr,"mtx: %d, GL_MODELVIEW_STACK_DEPTH = %d\n",mtx,depth);
		
		for (i=0; i<sz; i++) {
            w = widgets[i];
            if (mtx) { glPushMatrix(); }
            //glTranslatef(w->_rect.x,w->_rect.y,0);
            gfx_translate(w->_rect.x,w->_rect.y);
            w->_paint();
            gfx_translate(-w->_rect.x,-w->_rect.y);

            //glTranslatef(-w->_rect.x,-w->_rect.y,0);
            if (mtx) { glPopMatrix(); }
        }
    }
    
    virtual int event(Event *e_) {
		
#ifdef BUILD_SIXENSE
		if (e_->device > 1) { return 0; } // non first player devices can't use GUI
		if (e_->device == 1 && e_->type == EVT_MOTION
			&& (fabs((float)e_->dx)+fabs((float)e_->dy)) < 5) { return 0; } // sixense motion events must be SIGNIFICANT to count
#endif

        int r; int i;
        Event *e = e_;
        Event e2;
        r = Widget::event(e);
        if (r || e->type == EVT_CLICK) { return r; }
        
        Widget *prev,*cur,*w;
        
        // handle orientation events
        if (e->type == EVT_ORIENT) {
            r = 0;
            for (i=widgets.size()-1; i>=0; i--) {
                w = widgets[i];
                if (w->ignore) { continue; }
                int rr = w->event(e); if (rr) { r = rr; }
            }
            return r;
        }
        
        // handle focus / blur events
        prev = focus; cur = focus;
        if (e->type == EVT_DOWN) {
            cur = 0;
            for (i=widgets.size()-1; i>=0; i--) {
                w = widgets[i];
                if (w->ignore) { continue; }
                if (e_->x < w->_rect.x ||
                    e_->y < w->_rect.y ||
                    e_->x >= (w->_rect.x+w->_rect.w) ||
                    e_->y >= (w->_rect.y+w->_rect.h)) { continue; }
                cur = w;
                break;
            }
        }
        if (e->type == EVT_BLUR) { cur = 0; }
        if (cur != prev) {
            if (prev != 0) { e2.type = EVT_BLUR; prev->event(&e2); }
            if (cur != 0) { e2.type = EVT_FOCUS; e2.v = EVT_DOWN; cur->event(&e2); } 
            // NOTE: e2.v gives the Widget a hint if it was clicked on or tabbed to.
            // This is used by the TextInput widget to give proper behavior.
            focus = cur;
        }
    
        // handle enter / exit events
        prev = hover; cur = hover;
        if (e->type == EVT_MOTION || e->type == EVT_DOWN) {
            cur = 0;
            for (i=widgets.size()-1; i>=0; i--) {
                w = widgets[i];
                if (w->ignore) { continue; }
                if (e_->x < w->_rect.x ||
                    e_->y < w->_rect.y ||
                    e_->x >= (w->_rect.x+w->_rect.w) ||
                    e_->y >= (w->_rect.y+w->_rect.h)) { continue; }
                cur = w;
                break;
            }
        }
        if (e->type == EVT_EXIT) { cur = 0; }
        if (cur != prev) {
            if (prev != 0) { e2.type = EVT_EXIT; prev->event(&e2); }
            if (cur != 0) { e2.type = EVT_ENTER; cur->event(&e2); }
            hover = cur;
        }

        // done with those
        if (e->type == EVT_ENTER || e->type == EVT_EXIT || e->type == EVT_BLUR || e->type == EVT_FOCUS) { return 0; }

        // hover
        if (hover != 0 && hover != focus && (e->type == EVT_MOTION || e->type == EVT_WHEEL)) {
            w = hover;
            e2 = *e_;
            e2.x -= w->_rect.x;
            e2.y -= w->_rect.y;
            r = w->_event(&e2);
            if (r) { return r; }
            //return 0;
        }
        
        // events
        if (focus != 0) {
            w = focus;
            e2 = *e_;
            e2.x -= w->_rect.x;
            e2.y -= w->_rect.y;
            r = w->_event(&e2);
            if (r) { return r; }
            
            if (hover == focus && e->type == EVT_UP) {
                e2 = *e_;
                e2.type = EVT_CLICK;
                e2.x -= w->_rect.x;
                e2.y -= w->_rect.y;
                r = w->_event(&e2);
                if (r) { return r; }
            }
        }

        
        return 0;
    }
    virtual void loop() {
        Widget *w;
        unsigned int i;
        unsigned int sz = widgets.size();
  		for (i=0; i<sz; i++) {
            w = widgets[i];
            w->loop();
        }
    }        
        
    virtual ~Container() {
//         fprintf(stderr,"destroying the Container, has %d items\n",widgets.size());
        unsigned int i;
        for (i=0; i<widgets.size(); i++) {
            Widget *w = widgets[i];
            delete w;
//             widgets[i] = 0; //HACK: valgrind work-a-round
        }
    }

    
};

inline void Widget::_focus() {
        Event e; Widget *w;
        
        w = this; while (w) {
            e.type = EVT_BLUR; w->event(&e);
            w = w->parent;
        }
        
        w = this; while (w) {
            e.type = EVT_FOCUS; w->event(&e);
            Widget *_w = w;
            w = w->parent;
            if (w) {
                Container *c = (Container *)w;
                c->focus = _w;
            }
        }
    }
inline bool Widget::replace(const char *v, Widget *w) {
    Widget *_w = this->find(v);
    if (!_w) { return false; }
    Container *p = _w->parent;
    int x = _w->rect.x, y = _w->rect.y;
    w->b_col = _w->b_col;
    w->b_row = _w->b_row;
    w->b_colspan = _w->b_colspan;
    w->b_rowspan = _w->b_rowspan;
    w->b_x = _w->b_x;
    w->b_y = _w->b_y;
    p->remove(_w);
    p->add(w,x,y);
    return true;
};

inline MyRect Widget::get_abs_rect(bool inc_style){
    int x=0,y=0;
    Widget *w = this;
    while (w) {
        x += w->rect.x;
        y += w->rect.y;
        x += w->_rect.x;
        y += w->_rect.y;
        w = w->parent;
        if (w) {
            x -= w->view.x;
            y -= w->view.y;
        }
    }
    if (inc_style) {
        return MyRect(x-rect.x,y-rect.y,_rect.w,_rect.h);
    }
    return MyRect(x,y,rect.w,rect.h);
}

inline int t_sum(int *v, int a, int b) {
    int r =0;
    for (int i=a;i<b;i++) { r += v[i]; }
    return r;
}

inline int t_max(int a, int b) {
    if (a>b){ return a; }
    return b;
}
inline int t_min(int a, int b) {
    if (a<b) { return a; }
    return b;
}


static bool t_sort_by_colspan(Widget *a, Widget *b) {
    return a->b_colspan < b->b_colspan;
}
static bool t_sort_by_rowspan(Widget *a, Widget *b) {
    return a->b_rowspan < b->b_rowspan;
}

inline int t_div(int a,int b,int c) {
    int v=a/b,r=a%b;
    if (r != 0 && (c%b)<r) { v += 1; }
    return v;
}


struct Table : Container {
    int row,col;
    Widget *me;
    Table() : Container() {
        me = new Widget();
        this->add(me,0,0);
        me->ignore=1;
        row = -1; col = 0;
    }
    void tr() { row = row +1; col = 0; }
    void td(Widget *w, int colspan = 1, int rowspan = 1, int col = -1, int row = -1) {
        if (col != -1) { this->col = col; }
        if (row != -1) { this->row = row; }
        this->row = t_max(0,this->row);
        col = this->col; row = this->row;
        w->b_col = col; w->b_row = row;
        w->b_colspan = colspan;
        w->b_rowspan = rowspan;
        this->col = col + colspan;
        this->add(w,0,0);
    }
    

    void resize() { //int width=0,int height=0) {
        Container::resize();
    
        unsigned int i; Widget *w; int v,j,diff,ww,hh,n;
//         # figure out how many rows, cols we have & resize widgets
        int tcols=0,trows=0;
        for (i=0; i<widgets.size(); i++) {
            w = widgets[i];
            tcols = t_max(tcols,w->b_col+w->b_colspan);
            trows = t_max(trows,w->b_row+w->b_rowspan);
        }
        
//         widgets = self.widgets[:]
// 
//         # this widget represents the table widget :)
//         # this helps force s_height, s_width, etc
//         w = new Widget()
//         w->b_col=0;w->b_row = 0;
//         w->b_colspan=cols;w->b_rowspan =rows;
//         w->rect = Rect(0,0,s_width?s_width:width,s_height?s_height:height);
//         widgets.append(w)
        me->b_col = 0; me->b_row = 0;
        me->b_colspan = tcols; me->b_rowspan = trows;
        me->_rect = MyRect(0,0,STYLE(width,0),STYLE(height,0));

//         # set up the initial cell sizes
        int *cols = new int[tcols]; memset(cols,0,sizeof(int)*tcols);
        //fprintf(stderr,"%d\n",sizeof(cols));
        int *rows = new int[trows]; memset(rows,0,sizeof(int)*trows);
        //fprintf(stderr,"%d\n",sizeof(rows));

//         # size out our cols
        std::sort(widgets.begin(), widgets.end(), t_sort_by_colspan);
        for (i=0; i<widgets.size(); i++) {
            w = widgets[i];
            v = t_sum(cols,w->b_col,w->b_col+w->b_colspan);
            if (v >= w->_rect.w) { continue; }
            diff = w->_rect.w - v;
            for (j=w->b_col; j<w->b_col+w->b_colspan; j++) {
                cols[j] += t_div(diff,w->b_colspan,j);
            }
        }

//         # size out our rows
        std::sort(widgets.begin(), widgets.end(), t_sort_by_rowspan);
        for (i=0; i<widgets.size(); i++) {
            w = widgets[i];
            v = t_sum(rows,w->b_row,w->b_row+w->b_rowspan);
            if (v >= w->_rect.h) { continue; }
            diff = w->_rect.h - v;
            for (j=w->b_row; j<w->b_row+w->b_rowspan; j++) {
                rows[j] += t_div(diff,w->b_rowspan,j);
            }
        }

//         # calc positions
        ww = 0;
        int *cpos = new int[tcols]; memset(cpos,0,sizeof(int)*tcols);
        for (n=0; n<tcols; n++) {
            cpos[n] = ww;
            ww += cols[n];
        }

        hh = 0;
        int *rpos = new int[trows]; memset(rpos,0,sizeof(int)*trows);
        for (n=0; n<trows; n++) {
            rpos[n] = hh;
            hh += rows[n];
        }

//         # move widgets into position
        for (i=0; i<widgets.size(); i++) {
            w = widgets[i];
            w->_rect.x=cpos[w->b_col]; w->_rect.y = rpos[w->b_row];
            int w2 = t_sum(cols,w->b_col,w->b_col+w->b_colspan);
            int h2 = t_sum(rows,w->b_row,w->b_row+w->b_rowspan);
            if (w->_rect.w != w2 || w->_rect.h != h2) {
//                 w->resize(w2,h2);
                w->_resize(w2,h2);
            }
        }
    
        this->rect.w = ww;
        this->rect.h = hh;
		delete [] rows;
		delete [] cols;
		delete [] rpos;
		delete [] cpos;
    }
};


#define _gui_content this->widgets[0]

/*
 The box widget will expand to its container.
 The box widget's size will be exactly the width/height you give it,
 since it calculates sizes depending on style stuff.
 */
struct Box : Container {
    Box(Widget *content) :
        Container() {
        this->add(content,0,0);
    }
    
    Box(Widget *content, StyleSheet &s) :
        Container() {
        this->add(content,0,0);
        style = s;
    }
    
    virtual void resize() {
        Container::resize();
        Widget *w = _gui_content;
        STYLE_GET_SIZES;
        if (!STYLE(clip,0)) {
            this->rect.w = t_max(w->_rect.w,STYLE(width,0)-(left+right));
            this->rect.h = t_max(w->_rect.h,STYLE(height,0)-(top+bottom));
        } else {
            int v; 
            this->rect.w = w->_rect.w;
            if ((v=STYLE(width,0))!=0) {
                this->rect.w = STYLE(width,0)-(left+right);
            }
            this->rect.h = w->_rect.h;
            if ((v=STYLE(height,0))!=0) {
                this->rect.h = STYLE(height,0)-(top+bottom);
            }
        }
        this->_align();
    }
    virtual void paint() {
        if (STYLE(clip,0)) { this->_enable_clip(); }
        Container::paint();
        if (STYLE(clip,0)) { this->_disable_clip(); }
    }
    
    virtual void _align() {
        Widget *w = _gui_content;
        int dx = this->rect.w-w->_rect.w;
        int dy = this->rect.h-w->_rect.h;
        w->_rect.x = (STYLE(align,0)+1)*dx/2;
        w->_rect.y = (STYLE(valign,0)+1)*dy/2;
    }
    
    virtual void resize(int width,int height) {
        Widget::resize(width,height);
        this->rect.w = width;
        this->rect.h = height;
        this->_align();
    }
    
};

struct ScrollBox : Container {
    int state; int ay;
    
    ScrollBox(Widget *content, int w, int h) :
        Container()
        {
        this->add(content,0,0);
        this->style.width = w; this->style.height = h;
            state=0; ay = 0;
    }
    
    void resize() {
        Container::resize();
        
        this->rect.w = STYLE(width,0);
        this->rect.h = STYLE(height,0);
    }
    
    virtual void loop() {
        if (!state) {
            view.y = t_max(0,t_min(_gui_content->_rect.h-this->rect.h,view.y-ay));
            ay -= mysign(ay);
            ay -= mysign(ay);
        }
    }

        
    virtual int event(Event *e) {
        if (e->type == EVT_DOWN) {
            state = 0; ay = 0;
            Event _e = *e;
            _e.x += view.x;
            _e.y += view.y;
            Container::event(&_e);
            return 0;
        }
        if (e->type == EVT_WHEEL) {
            view.y = t_max(0,t_min(_gui_content->_rect.h-this->rect.h,view.y+e->dy*30));
            return 0;
        }
            
        if (e->type == EVT_MOTION && e->b != 0 && style.focus != 0) {
            view.x = t_max(0,t_min(_gui_content->_rect.w-this->rect.w,view.x-e->dx));
            view.y = t_max(0,t_min(_gui_content->_rect.h-this->rect.h,view.y-e->dy));
            if (e->dy) { ay = e->dy; }
            state = 1;
            return 0;
        }
        if (e->type == EVT_MOTION) {
            int skip = (e->x < 0 || e->y < 0 || e->x >= rect.w || e->y >= rect.h);
            Event _e = *e;
            _e.x += view.x;
            _e.y += view.y;
            if (skip) {
                _e.x = -65536;
                _e.y = -65536;
            }
            return Container::event(&_e);
        }
        if (e->type == EVT_UP) {
            if (state) { state=0; return 0; }
            Event _e = *e;
            _e.x += view.x;
            _e.y += view.y;
            return Container::event(&_e);
        }
        
        return Container::event(e);
    }
    
    virtual void paint() {
        this->_enable_clip();
        
//         glPushMatrix();
        //glTranslatef(-view.x,-view.y,0);
        gfx_translate(-view.x,-view.y);

        Container::paint();
        gfx_translate(view.x,view.y);
//         glPopMatrix();
        
        this->_disable_clip();
    }
};


struct Label : Widget {
    Font fnt;
    char *txt;
    Label(Font f, const char *s) : Widget(), fnt(f) {
        this->txt = strdup(s);
    }

    virtual void paint() {
         unsigned int c = STYLE(color,0xffffff);
         glColor4ub(((c>>16)&0xff),((c>>8)&0xff),((c>>0)&0xff),0xff);
         fnt.draw(txt,0,0,this->rect.w,STYLE(align,-1));
         glColor4ub(0xff,0xff,0xff,0xff);
    }
    
    virtual void resize() {
        this->rect.w = STYLE(width,(int)fnt.get_width(txt));
        this->rect.h = (int)fnt.get_height(txt,this->rect.w);
    }
    
    virtual int event(Event *e) {
        return Widget::event(e);
    }
    
    virtual ~Label() {
        free(txt);
    }
};


#include <vector>

struct TextInputChange {
    int p;
    int pos;
    std::string a;
    std::string b;
    TextInputChange() : p(0), pos(0), a(""), b("") { }
    TextInputChange(int _p, int _pos, std::string _a, std::string _b) : p(_p), pos(_pos), a(_a), b(_b) { }
    void print(const char *ctx, int n) {
//        fprintf(stderr,"%s %d: %d %d '%s' '%s'\n",ctx,n,p,pos,a.c_str(),b.c_str());
    }
};

struct TextInput : Widget {
    Font fnt;
    std::string txt; std::string txt2;
    std::vector<TextInputChange> undo; int undo_n;
    int _pos,pos; int cursor_w; int cx;
    unsigned int maxlen; int passwd; int reveal; int wrap;
    TextInput(Font f, const char *s, unsigned int _maxlen=255) : Widget(), fnt(f) {
        _maxlen = (int)mymax(_maxlen,strlen(s));
        this->maxlen = _maxlen;
        undo_n = 0;
        passwd=0;
        reveal=0;
        wrap = 0;
        _pos = pos = 0;
        view.x = view.y = 0;
        cursor_w = 2;
        cx = -1;
        
        txt = std::string(s); // NOTE: you can inject '\n' in this way
        _mktxt2();
        
        this->style.loop();
        this->resize();

        _adjust(false,false);
        
        //fprintf(stderr,"view.x: %d view.y: %d\n",(int)view.x,(int)view.y);
        
    }
    
    std::string _filter(std::string ct) {
        for (int i=0; i<ct.size(); i++) {
            if (wrap == 0 && ct[i] == '\n') { ct.replace(i,1,1,' '); }
            if (ct[i] != '\n' && (ct[i] < 32 || ct[i] >= 127)) { ct.replace(i,1,1,'?'); }
        }
        return ct;
    }
    
    virtual const char *get_value_s() { return txt.c_str(); }
    virtual void set_value_s(const char *s) {
        _change(0,txt.size(),std::string(s));
    }

    void _apply(int p, int n, std::string v) {
        std::string a = txt.substr(0,p);
        std::string b = txt.substr(p+n,txt.size()-(p+n));
        txt = a+v+b;
        _mktxt2();
    }
    
    int _min(int a, int b) {
        return a<b?a:b;
    }
    int _max(int a, int b) {
        return a>b?a:b;
    }
    int _abs(int v) {
        return v>0?v:-v;
    }
    
    void _change(int p, int n, std::string b) {
        if (p < 0 || n < 0) { return; }
        if ((p+n) > txt.size()) { n = txt.size()-p; }
        if (n < 0) { return; }

        int size = txt.size() - n + b.size();
        if (size > maxlen) {
            int bsize = b.size() - (size-maxlen);
            if (bsize < 0) { return; }
            b.resize(bsize);
        }
        b = _filter(b);
        
        TextInputChange ch = TextInputChange(pos,p,txt.substr(p,n),b);
        ch.print("_change",undo_n);
        _apply(ch.pos,ch.a.size(),ch.b);
        _pos = pos = ch.pos + ch.b.size();
        
        undo.resize(undo_n);
        undo.push_back(ch);
        undo_n += 1;
        if (undo_n > 255) {
            undo.erase(undo.begin());
            undo_n -= 1;
        }
    }
    
    void _undo() {
        if (!undo_n) { return; }
        undo_n -= 1;
        TextInputChange ch = undo[undo_n];
        ch.print("_undo",undo_n);
        _apply(ch.pos,ch.b.size(),ch.a);
        _pos = pos = ch.p; // ch.pos + ch.a.size();
    }
    
    void _redo() {
        if (undo_n >= undo.size()) { return; }
        TextInputChange ch = undo[undo_n];
        ch.print("_redo",undo_n);
        _apply(ch.pos,ch.a.size(),ch.b);
        _pos = pos = ch.pos + ch.b.size();
        undo_n += 1;
        
    }
    
    void _adjust(bool _cx=true, bool cursor=true) {
        if (_cx) { cx = -1; }
        int ww = this->wrap?this->rect.w-cursor_w:0;
        int w = (int)fnt.get_width(txt2.c_str(),ww);
        int h = (int)fnt.get_height(txt2.c_str(),ww);
        if (cursor) {
            GfxRect r = fnt.crect(txt2.c_str(),pos,ww);
            r.w=cursor_w;
            if (r.x < view.x) { view.x = r.x; }
            if (r.x+r.w > view.x+rect.w) { view.x = r.x+r.w-rect.w; }
            if (r.y < view.y) { view.y = r.y; }
            if (r.y+r.h > view.y+rect.h) { view.y = r.y+r.h-rect.h; }
        }
        view.x = _max(0,_min(w-rect.w+cursor_w,view.x));
        view.y = _max(0,_min(h-rect.h,view.y));
    }
    
    void _click(int x,int y, bool both=true) {
        int ww = this->wrap?this->rect.w-cursor_w:0;
        pos = fnt.cnear(txt2.c_str(),x,y,ww);
        if (both) { _pos = pos; }
    }
    
    void _mktxt2() {
        txt2 = txt;
        if (passwd) {
            int l = txt2.size();
            if (reveal && l > 0) {
                txt2 = std::string(l-1,'*') + std::string(1,txt[l-1]);
            } else {
                txt2 = std::string(l,'*');
            }
        }
    }
    
    virtual void paint() {
        _mktxt2(); // it'd be nice to only do this as needed
        
        this->_enable_clip();
        int ww = this->wrap?this->rect.w-cursor_w:0;
        
        gfx_translate(-view.x,-view.y);
        if (style.focus) {
            GfxRect r = fnt.crect(txt2.c_str(),pos,ww);
            r.w=cursor_w;
            if (_pos != pos) {
                GfxRect r1 = fnt.crect(txt2.c_str(),_min(_pos,pos),ww);
                GfxRect r2 = fnt.crect(txt2.c_str(),_max(_pos,pos),ww);
                r2.w = 0;
                int c = 0x55aaff;
                if (r1.y == r2.y) {
                    gfx_fill(c,r1.x,r1.y,r2.x+r2.w-r1.x,r1.h);
                } else {
                    gfx_fill(c,r1.x,r1.y,ww-r1.x,r1.h);
                    gfx_fill(c,0,r1.y+r1.h,ww,r2.y-(r1.y+r1.h));
                    gfx_fill(c,0,r2.y,r2.x+r2.w,r2.h);
                }
            }
            fnt.draw(txt2.c_str(),0,0,ww);
            gfx_fill(0xffffff,r.x,r.y,r.w,r.h);
        } else {
            fnt.draw(txt2.c_str(),0,0,ww);
        }
        gfx_translate(view.x,view.y);
        
        this->_disable_clip();
    }
    
    virtual void resize() {
        int ww = this->wrap?this->rect.w:0;
        this->rect.h = STYLE(height,(int)fnt.get_height(this->txt.c_str(),ww));
        this->rect.w = STYLE(width,(int)mymax(this->rect.h,fnt.get_width(txt.c_str())));
    }
    
    virtual int event(Event *e) {
        int r = Widget::event(e);
        if (r) { return r; }
        //unsigned int l = txt.size();
        
        
        // by handling click we can re-show the keyboard on all re-clicks
        if (e->type == EVT_CLICK) {
            reveal=0;
            driver_keyboard_show();
            MyRect r = get_abs_rect(true);
            gfx_focus(GfxRect(r.x,r.y,r.w,r.h));
            return 0;
        }
        
        if (e->type == EVT_FOCUS) {
            fprintf(stderr,"evt_focus\n");
            if (!wrap) {
                _pos = 0; pos = txt.size();
            }
            _adjust(false,e->v != EVT_DOWN);
            return 0;
        }
        if (e->type == EVT_BLUR) {
            reveal=0;
            driver_keyboard_hide();
            gfx_blur();
            
            if (!wrap) {
                view.x = 0; view.y = 0;
            }
            _adjust(false,false);
            return 0;
        }
        if (e->type == EVT_DOWN) {
            fprintf(stderr,"evt_down\n");
            _click(view.x + e->x,view.y + e->y);
            _adjust();
            return 0;
        }
        if (e->type == EVT_MOTION && style.focus != 0 && e->b != 0) {
            _click(view.x + e->x,view.y + e->y,false);
            _adjust();
            return 0;
        }
        if (e->type == EVT_UP) {
            _click(view.x + e->x,view.y + e->y,false);
            _adjust();
            return 0;
        }
        if (e->type == EVT_WHEEL) {
            view.y += e->dy * fnt.get_height();
            _adjust(false,false);
            return 0;
        }
        if (e->type == EVT_KEYDOWN) {
            reveal=0;
            if (e->v == KEY_TAB) {
                return 0;
            }
            if (e->v == KEY_BACKSPACE) {
                if (_pos != pos) {
                    _change(_min(_pos,pos),_abs(_pos-pos),std::string(""));
                } else if (pos) {
                    _change(pos-1,1,std::string(""));
                }
                _adjust();
                return 0;
            }
            if (e->v == KEY_DELETE) {
                if (_pos != pos) {
                    _change(_min(_pos,pos),_abs(_pos-pos),std::string(""));
                } else if (1) {
                    _change(pos,1,std::string(""));
                }
                _adjust();
                return 0;
            }            
            if (e->v == KEY_LEFT) {
                if (pos) { pos -= 1;}
                if ((e->b&KEYM_SHIFT) == 0) {_pos = pos;}
                _adjust();
                return 0;
            }
            if (e->v == KEY_RIGHT) {
                if (pos < txt.size()) { pos += 1;}
                if ((e->b&KEYM_SHIFT) == 0) {_pos = pos;}
                _adjust();
                return 0;
            }
            if (e->v == KEY_UP) {
                int ww = this->wrap?this->rect.w-cursor_w:0;
                GfxRect r = fnt.crect(txt2.c_str(),pos,ww);
                if (cx == -1) { cx = r.x+r.w/2; }
                _click(cx,r.y-r.h/2,(e->b&KEYM_SHIFT) == 0);
                _adjust(false,true);
                return 0;
            }
            if (e->v == KEY_DOWN) {
                int ww = this->wrap?this->rect.w-cursor_w:0;
                GfxRect r = fnt.crect(txt2.c_str(),pos,ww);
                if (cx == -1) { cx = r.x+r.w/2; }
                _click(cx,r.y+r.h+r.h/2,(e->b&KEYM_SHIFT) == 0);
                _adjust(false,true);
                return 0;
            }
            if (e->v == KEY_HOME) {
                int ww = this->wrap?this->rect.w-cursor_w:0;
                GfxRect r = fnt.crect(txt2.c_str(),pos,ww);
                _click(0,r.y+r.h/2,(e->b&KEYM_SHIFT) == 0);
                _adjust();
                return 0;
            }
            if (e->v == KEY_END) {
                int ww = this->wrap?this->rect.w-cursor_w:0;
                int w = (int)fnt.get_width(txt2.c_str(),ww);
                GfxRect r = fnt.crect(txt2.c_str(),pos,ww);
                _click(w,r.y+r.h/2,(e->b&KEYM_SHIFT) == 0);
                _adjust();
                return 0;
            }
            
            if (e->v == KEY_RETURN) {
                //fprintf(stderr,"entered: %s\n",txt);
                // emit on-change or something ..
                if (!wrap) {
                    driver_keyboard_hide();
                    return onchange;
                }
                e->c = '\n';
            }
            if (e->v == KEY_z && (e->b&(KEYM_ALT|KEYM_CTRL|KEYM_OPT))!=0 && (e->b&KEYM_SHIFT)==0) {
                _undo();
                _adjust();
                return 0;
            }
            if (e->v == KEY_z && (e->b&(KEYM_ALT|KEYM_CTRL|KEYM_OPT))!=0 && (e->b&KEYM_SHIFT)!=0) {
                _redo();
                _adjust();
                return 0;
            }
            
            if (e->v == KEY_x && (e->b&(KEYM_ALT|KEYM_CTRL|KEYM_OPT))!=0) {
                if (!this->passwd) {
                    driver_clipboard_set_text(std::string(txt).substr(_min(_pos,pos),_abs(_pos-pos)));
                }
                _change(_min(_pos,pos),_abs(_pos-pos),std::string(""));
                _adjust();
                return 0;
            }
            if (e->v == KEY_c && (e->b&(KEYM_ALT|KEYM_CTRL|KEYM_OPT))!=0) {
                if (!this->passwd) {
                    driver_clipboard_set_text(std::string(txt).substr(_min(_pos,pos),_abs(_pos-pos)));
                }
                return 0;
            }
            if (e->v == KEY_v && (e->b&(KEYM_ALT|KEYM_CTRL|KEYM_OPT))!=0) {
                std::string ct = driver_clipboard_get_text();
                _change(_min(_pos,pos),_abs(_pos-pos),ct);
                _adjust();
                return 0;
            }

            #ifdef BUILD_IGALCON2
            reveal=1;
            #endif
            if (e->c == '\n' || (e->c >= 32 && e->c < 127)) {
                _change(_min(_pos,pos),_abs(_pos-pos),std::string(1,e->c));
                _adjust();
            }
        }

        return 0;
    }
    
//    virtual ~TextInput() {
//        free(txt);
//    }
};


#endif

