#include "cuz_var.h"

#include "cuz_json.h"

#ifdef JSON_TEST

#include <sys/types.h>
#include <sys/stat.h>
const char *driver_data_fname(const char *_fname) {
    static char fname[512];
    sprintf(fname,"./%s",_fname);
    return fname;
}

int _driver_fsize(const char *fname) {
    struct stat stbuf;
	if (stat(fname, &stbuf)) { return 0; }
    int fsize = stbuf.st_size;
    return fsize;
}


std::string driver_data_load(const char *_fname) {
    const char *fname = driver_data_fname(_fname);
    
    int fsize = _driver_fsize(fname);

    FILE* f = fopen(fname,"rb");
    char* mem = new char[fsize];
	if (f) {
		fread(mem, fsize, 1, f);
		fclose(f);
	}
    
    std::string r = std::string(mem,fsize);
    delete [] mem;
    return r;
}
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <locale.h>

#include "JSON_parser.h"

static int _json_callback(void* ctx, int type, const JSON_value* value);

struct json_ctx {
    std::vector<var*> stack;
    var *cur;
    var &top;
    std::string key;
    
    json_ctx(var &_top) : stack(), cur(0),top(_top),key() { }
    
    void begin(var v) {
        if (!stack.size()) {
            top = v;
            cur = &top;
            stack.push_back(cur);
            return;
        }
        cur = push(v);
        stack.push_back(cur);
    }
    void end() {
        cur = 0;
        stack.pop_back();
        if (stack.size()) {
            cur = stack[stack.size()-1];
        }
    }
    void set_key(const char *s) {
        key = std::string(s);
    }
    var *push(var v) {
        if (!cur) { return 0; }
        if (cur->_type == VAR_LIST) {
            cur->append(v);
            return &((*cur)[cur->__len__()-1]);
        }
        if (cur->_type == VAR_DICT) {
            (*cur)[key] = v;
            return &((*cur)[key]);
        }
        return 0;
    }
};

int json_decode(var &_top,const char *data, int len) {
    int result = 0;
    JSON_config config;
    struct JSON_parser_struct* jc = NULL;
    
    init_JSON_config(&config);
    config.depth                  = 16;
    config.callback               = &_json_callback;
    config.allow_comments         = 1;
    config.handle_floats_manually = 0;
    json_ctx ctx(_top);
    config.callback_ctx = &ctx;
//     setlocale(LC_ALL,"C");
    jc = new_JSON_parser(&config);
    
    for (int n=0; n<len; n++) {
        int next_char = data[n];
        if (next_char <= 0) {
            break;
        }
        if (!JSON_parser_char(jc, next_char)) {
            fprintf(stderr, "JSON_parser_char: syntax error, byte %d\n", n);
            result = 1;
            goto done;
        }
    }
    if (!JSON_parser_done(jc)) {
        fprintf(stderr, "JSON_parser_end: syntax error\n");
        result = 1;
        goto done;
    }
    
done:
    delete_JSON_parser(jc);
    return result;
}

// int total=0;

static int _json_callback(void* _ctx, int type, const JSON_value* value) {
    json_ctx *ctx = (json_ctx*)_ctx;
    switch (type) {
        case JSON_T_ARRAY_BEGIN: ctx->begin(var(VAR_LIST,0)); break;
        case JSON_T_ARRAY_END: ctx->end(); break;
        case JSON_T_OBJECT_BEGIN: ctx->begin(var(VAR_DICT,0)); break;
        case JSON_T_OBJECT_END: ctx->end(); break;
        case JSON_T_KEY: ctx->set_key(value->vu.str.value); break;
        case JSON_T_INTEGER: ctx->push((var_number)value->vu.integer_value); break;
        case JSON_T_FLOAT: ctx->push((var_number)value->vu.float_value); break;
        case JSON_T_NULL: ctx->push(None); break;
        case JSON_T_TRUE: ctx->push(1); break;
        case JSON_T_FALSE: ctx->push(0); break;
        case JSON_T_STRING: ctx->push(value->vu.str.value); break;
        default: assert(0); break;
    }
    return 1;
}

#include <sstream>
void _json_encode(var &v,std::stringstream *ss) ;

void _json_encode_number(var &v, std::stringstream *ss) {
    *ss << v._number;
}
void _json_encode_string(std::string &_v, std::stringstream *ss) {
    const char *v = _v.c_str();
    ss->put('"');
    for (char c=*v; c!=0; c=*(++v)) {
        switch (c) {
            case '\"': ss->write("\\\"",2); break;
            case '\\': ss->write("\\\\",2); break;
//             case '/': ss->write("\\/",2); break;
            case '\b': ss->write("\\b",2); break;
            case '\f': ss->write("\\f",2); break;
            case '\r': ss->write("\\r",2); break;
            case '\n': ss->write("\\n",2); break;
            case '\t': ss->write("\\t",2); break;
            default:
                if (c > 0 && c <= 0x1F) { 
                    char tmp[8];
                    sprintf(tmp,"\\u%04x",c);
                    ss->write(tmp,strlen(tmp));
                    break;
                }
                ss->put(c);
                break;
        }
    }
    ss->put('"');
}

void _json_encode_list(var &e, std::stringstream *ss) {
    int n = 0;
    ss->put('[');
    for_in(v,e) {
        if (n) { ss->put(','); } n += 1;
        _json_encode(v,ss);
    }
    ss->put(']');
}

void _json_encode_dict(var &e, std::stringstream *ss) {
    int n = 0;
    ss->put('{');
    pair_in(k,v,e) {
        if (n) { ss->put(','); } n += 1;
        _json_encode_string(k._string,ss);
        ss->put(':');
        _json_encode(v,ss);
    }
/*    std::map<std::string,var>::iterator dict_it,dict_end;
    dict_it = e._dict->begin(); dict_end = e._dict->end();
    for (; dict_it != dict_end; dict_it++) {
        if (n) { ss->put(','); } n += 1;
        _json_encode_string((*dict_it).first,ss);
        ss->put(':');
        _json_encode((*dict_it).second,ss);
    }*/
    ss->put('}');
}

void _json_encode(var &v,std::stringstream *ss) {
    switch(v._type) {
        case VAR_NUMBER: _json_encode_number(v,ss); return; break;
        case VAR_STRING: _json_encode_string(v._string,ss); return; break;
        case VAR_LIST: _json_encode_list(v,ss); return; break;
        case VAR_DICT: _json_encode_dict(v,ss); return; break;
    }
    ss->write("null",4);
}

std::string json_encode(var &v) {
    std::stringstream ss;
    _json_encode(v,&ss);
    return ss.str();
}

#ifdef JSON_TEST

int main(int argc, char *argv[]) {

    var js;
    std::string jdata = driver_data_load("test.json");
    json_decode(js,jdata.c_str(),jdata.size());
//     print(js["blast-0"]["texture"]);
    
    var goods = var(VAR_DICT,0);
    for (int i=0; i<256; i++) {
        goods[var_str(i)] = js;
    }
    std::string jout = json_encode(goods);
//     print(jout);
    
//     std::string jout = json_encode(js);
    
//     assert(tp_abs(jdata.size()-jout.size())<2);
    
//     fprintf(stderr,"%d + %d = %d\n",total,hits,total+hits);
    
//     print(jout);
//     var js2;
//     json_decode(js2,jout.c_str(),jout.size());
//     print(js2["blast-0"]["texture"]);
    
}

#endif
//12 bytes, with -O3 -fno-exceptions
//==2428== malloc/free: 13,949 allocs, 13,949 frees, 623,568 bytes allocated.
//==2429== I   refs:      29,785,162

//==25475== I   refs:      159,289,146

// no optimizations, w/o counted_ptr
//==31616== malloc/free: 13,069 allocs, 13,069 frees, 626,322 bytes allocated.
//==31609== I   refs:      55,503,148
