#include "cuz_driver.h"
#include "cuz_mixer.h"

#include "cuz_json.h"
struct _MixerData {
    
    var sample_json; // the JSON file data
    std::map<std::string,int> sample_map; // a map from a name to an id

    var stream_json; // the JSON file data
    std::map<std::string,int> stream_map; // a map from a name to an id
};


static _MixerData *_mixer = 0;

void mixer_init() {

    _mixer = new _MixerData();

    std::string json;
    
    json = driver_data_load("data.json");
    json_decode(_mixer->sample_json,json.c_str(),json.size());
    
    json = driver_data_load("data.json");
    json_decode(_mixer->stream_json,json.c_str(),json.size());
}

void mixer_deinit() {

    delete _mixer;
    _mixer = 0;
}

int mixer_sample_load(const char *name) {
    var &js = _mixer->sample_json;
    
    if (_mixer->sample_map.find(name) == _mixer->sample_map.end()) {
        fprintf(stderr,"trying load sample %s\n",name);
        if (!js.has(name)) {
            fprintf(stderr,"unable to find sample %s\n",name);
            _mixer->sample_map[name] = 0;
            return 0;
        }
        fprintf(stderr,"doing load sample %s\n",name);
        _mixer->sample_map[name] = driver_audio_sample_load(driver_data_fname(js[name]["fname"].c_str()));
    }
    
    return _mixer->sample_map[name];    
    
}

void mixer_sample_play(int v, float vol,float pan) {
    driver_audio_sample_play(v,vol,pan);
}

int mixer_stream_load(const char *name) {
    var &js = _mixer->stream_json;
    
    if (_mixer->stream_map.find(name) == _mixer->stream_map.end()) {
        fprintf(stderr,"trying load stream %s\n",name);
        if (!js.has(name)) {
            fprintf(stderr,"unable to find stream %s\n",name);
            _mixer->stream_map[name] = 0;
            return 0;
        }
        fprintf(stderr,"doing load stream %s\n",name);
        _mixer->stream_map[name] = driver_audio_stream_load(driver_data_fname(js[name]["fname"].c_str()));
    }
    
    return _mixer->stream_map[name];
}

void mixer_stream_play(int v, float vol) {
    driver_audio_stream_play(v,vol);
}

void mixer_stream_stop(int v) {
    driver_audio_stream_stop(v);
}

void mixer_stream_volume(int v, float vol) {
    driver_audio_stream_volume(v,vol);
}
