#ifndef CUZ_VAR_H
#define CUZ_VAR_H

#include <stdarg.h>
#include <assert.h>
#include <stdio.h>

#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <algorithm>

#define print var_print
// #define str var_str
#define len var_len
#define None _var_none()
#define Error _var_error()
#define list var_list
#define dict var_dict
#define for_in var_for_in
#define pair_in var_pair_in

inline std::string &var_get_errstr() {
    static std::string var_errstr;
    return var_errstr;
}

#define VAR_THROW(msg,ret) { fprintf(stderr,"%s:%d: %s\n",__FILE__,__LINE__,msg); var_get_errstr() = std::string(msg); return ret; }

inline void check_error(std::string val) {
    assert(var_get_errstr().find(val) != std::string::npos);
    var_get_errstr() = std::string("");
}
#define VAR_ASSERT(val,msg,ret) if (!(val)) { VAR_THROW(msg,ret); }

enum { VAR_ERROR=-1,VAR_NONE,VAR_NUMBER,VAR_STRING,VAR_LIST,VAR_DICT };
struct _var_none { _var_none() { } };
struct _var_error { _var_error() { } };
#define var_map std::map
#define var_vector std::vector
typedef float var_number;

inline var_number var_max(var_number a, var_number b) {
    return a>b?a:b;
}
inline var_number var_min(var_number a, var_number b) {
    return a<b?a:b;
}
inline var_number var_abs(var_number a) {
    return a>0?a:-a;
}


struct var {
    int _type;
    union {
        int _union;
        var_number _number;
        var_vector<var> *_list;
        var_map<std::string,var> *_dict;
    };
    std::string _string;
    
    // constructors ////////////////////////////////////////////////////////////
    
    var() {
        _init();
        _type = VAR_NONE;
    }
    var(const var &value) {
        _init();
//         fprintf(stderr,"var(const var &value): %d = %s\n",value._type,value.__str__().c_str());
        _type = value._type;
        if (_type == VAR_NUMBER) { _number = value._number; }
//         _list = value._list;
//         _dict = value._dict;
        if (_type== VAR_LIST) {_chlist(value._list);}
        if (_type == VAR_DICT) { _chdict(value._dict);}
        _string = value._string;
        
        
    }
    var(const _var_error value) {
        _init();
        _type = VAR_ERROR;
    }
    var(const _var_none value) {
        _init();
        _type = VAR_NONE;
    }
    
    var(const float value) {
        _init();
        _type = VAR_NUMBER;
        _number = (var_number)value;
    }
    var(const int value) {
        _init();
        _type = VAR_NUMBER;
        _number = (var_number)value;
    }
    var(const double value) {
        _init();
        _type = VAR_NUMBER;
        _number = (var_number)value;
    }
    var(const char *value) {
        _init();
        _type = VAR_STRING;
        _string = std::string(value);
    }
    var(std::string value) {
        _init();
        _type = VAR_STRING;
        _string = value;
    }
    
    var(int type,int total,...) {
        _init();
        va_list args;
        va_start(args,total);
        if (type == VAR_LIST) {
            _type = VAR_LIST;
//             fprintf(stderr,"=var list..begin\n");
            if (!total) { va_end(args); return; }
            _listcheck(1);
            (*_list).reserve(total);
            for (int i=0; i<total; i++) {
                (*_list).push_back(*va_arg(args,var*));
            }
//             fprintf(stderr,"=var list..end\n");
            va_end(args);
            return;
        }
        if (type == VAR_DICT) {
            _type = VAR_DICT;
//             fprintf(stderr,"=var dict..begin\n");
            if (!total) { va_end(args); return; }
            _dictcheck(1);
            for (int i=0; i<total; i+=2) {
//                 fprintf(stderr,"tick..\n");
//                 _dict.insert(std::pair<var,var>(*e,*(e+1) ) );
//                 _dict.insert(std::make_pair(*e,*(e+1) ) );
                var *k = va_arg(args,var*);
                (*_dict)[k->__str__()]=*va_arg(args,var*);
            }
//             fprintf(stderr,"=var dict..end\n");
            va_end(args); 
            return;
        }
        _type = VAR_ERROR;
        va_end(args);
        return;
    }
    
    // casts ///////////////////////////////////////////////////////////////////
    
    operator double() {
        if (_type != VAR_NUMBER) { return 0; }
        return (double)_number;
    }
    operator float() {
        if (_type != VAR_NUMBER) { return 0; }
        return (float)_number;
    }
    operator int() {
        if (_type != VAR_NUMBER) { return 0; }
        return (int)_number;
    }
    operator std::string() {
//         if (_type != VAR_STRING) { return std::string(""); }
        return __str__();
    }
    
    // destructor //////////////////////////////////////////////////////////////
    
    ~var() {
        _destroy();
    }
    
    // internal / private stuff ////////////////////////////////////////////////
    void _init() {
        _type = VAR_ERROR;
        _string = std::string();
        _union = 0;
/*        _number = 0;
        _list = 0;
        _dict = 0;*/
//         _list = var_vector<var>();
//         _dict = var_map<std::string,var>();
    }
    void _chlist(var_vector<var> *v) {
        if (_list != 0 && v != 0) { 
            *_list = *v;
            return;
        }
        if (_list) { delete _list; _list = 0; }
        if (v) {
            _list = new var_vector<var>();
            *_list = *v;
        }
    }
    
    void _chdict(var_map<std::string,var> *v) {
        if (_dict != 0 && v != 0) { 
            *_dict = *v;
            return;
        }
        if (_dict) { delete _dict; _dict = 0; }
        if (v) {
            _dict = new var_map<std::string,var>();
            *_dict = *v;
        }
    }
    int _listcheck(int mk) {
        if (_type != VAR_LIST) { return 0; }
        if (_list) { return 1; }
        if (mk) {
            _list =  new var_vector<var>;
        }
        return _list != 0;
    }
    
    int _dictcheck(int mk) {
        if (_type != VAR_DICT) { return 0; }
        if (_dict) { return 1; }
        if (mk) {
            _dict =  new var_map<std::string,var>;
        }
        return _dict != 0;
    }
    
    void _destroy() {
        switch(_type) {
            case VAR_NUMBER: _number = 0; break;
            // NOTE: clearing the string causes a lot more memory reallocation in json_encode
            case VAR_STRING: /*_string.clear();*/ break;
            case VAR_LIST: _chlist(0); break;
            case VAR_DICT: _chdict(0); break;
//             case VAR_LIST: _list.clear(); break;
//             case VAR_DICT: _dict.clear(); break;
        }
    }

    
    // assignment operators ////////////////////////////////////////////////////
    void operator=(const var &value) {
        if (&value == this) { return; }
//         fprintf(stderr,"operator=(const var &value): %d = %s\n",value._type,value.__str__().c_str());
        if (_type == VAR_LIST && value._type != _type) { _chlist(0);}
        if (_type == VAR_DICT && value._type != _type) { _chdict(0);}
        if (value._type != _type) { _union = 0; }
        _type = value._type;
//         _list = value._list;
//         _dict = value._dict;
        if (_type == VAR_NUMBER) { _number = value._number; }
        if (_type== VAR_LIST) { _chlist(value._list);}
        if (_type == VAR_DICT) { _chdict(value._dict);}
        _string = value._string;
    }
    
    void operator=(const _var_none value) {
        _destroy();
        _type = VAR_NONE;
    }
    void operator=(const float value) {
        _destroy();
        _type = VAR_NUMBER;
        _number = (var_number)value;
    }
    void operator=(const double value) {
        _destroy();
        _type = VAR_NUMBER;
        _number = (var_number)value;
    }
    void operator=(const int value) {
        _destroy();
        _type = VAR_NUMBER;
        _number = (var_number)value;
    };
    void operator=(const char *value) {
        _destroy();
        _type = VAR_STRING;
        _string = std::string(value);
    }
    void operator=(std::string value) {
        _destroy();
        _type = VAR_STRING;
        _string = value;
    }

    // boolean operators ///////////////////////////////////////////////////////
    
    bool operator<(const var &b) const {
        if (_type != b._type) { return _type < b._type; }
        switch(_type) {
            case VAR_NONE:
                return false;
                break;
            case VAR_NUMBER:
                return _number < b._number;
                break;
            case VAR_STRING:
                return _string < b._string;
                break;
            case VAR_LIST:
                return this < &b;
                break;
            case VAR_DICT:
                return this < &b;
                break;
        }
        return false;
    }
    
    bool operator==(const var &b) const {
        if (_type != b._type) { return false; }
        switch(_type) {
            case VAR_NONE:
                return true;
                break;
            case VAR_NUMBER:
                return _number == b._number;
                break;
            case VAR_STRING:
                return _string == b._string;
                break;
            case VAR_LIST:
                return this == &b;
                break;
            case VAR_DICT:
                return this == &b;
                break;
        }
        return false;
    }
    
    bool operator!=(const var &b) const {
        return ! (*this == b);
    }
    bool operator==(const double &v) const {
        return (_type == VAR_NUMBER && _number == v);
    }
    bool operator==(const float &v) const {
        return (_type == VAR_NUMBER && _number == (var_number)v);
    }
    bool operator==(const int &v) const {
        return (_type == VAR_NUMBER && _number == (var_number)v);
    }
    bool operator==(const std::string &v) const {
        return (_type == VAR_STRING && _string == v);
    }
    bool operator==(const char *v) const {
        return (_type == VAR_STRING && _string == std::string(v));
    }

    // arithmetic operators ////////////////////////////////////////////////////
    var operator+(const var &b) const {
        VAR_ASSERT(_type == b._type,"TypeError",Error);
        switch(_type) {
            case VAR_STRING:
                return var(_string + b._string);
                break;
            case VAR_NUMBER:
                return var(_number + b._number);
                break;
            case VAR_LIST:
                {
                var tmp = var(VAR_LIST,0);
                tmp._listcheck(1);
                tmp._list->reserve((_list?_list->size():0) + (b._list?b._list->size():0));
                if (_list) {
                    tmp._list->insert(tmp._list->end(),_list->begin(),_list->end());
                }
                if (b._list) {
                    tmp._list->insert(tmp._list->end(),b._list->begin(),b._list->end());
                }
                return tmp;
                
                }
                break;
        }
        VAR_THROW("UnsupportedError",Error);
    }
    
    var operator*(const var &b) const {
        VAR_ASSERT(_type == b._type,"TypeError",Error);
        switch(_type) {
            case VAR_NUMBER:
                return var(_number * b._number);
                break;
        }
        VAR_THROW("UnsupportedError",Error);
    }
    var operator/(const var &b) const {
        VAR_ASSERT(_type == b._type,"TypeError",Error);
        switch(_type) {
            case VAR_NUMBER:
                return var(_number / b._number);
                break;
        }
        VAR_THROW("UnsupportedError",Error);
    }
    var operator-(const var &b) const {
        VAR_ASSERT(_type == b._type,"TypeError",Error);
        switch(_type) {
            case VAR_NUMBER:
                return var(_number - b._number);
                break;
        }
        VAR_THROW("UnsupportedError",Error);
    }
    void operator+=(const var &value) {
        *this = *this + value;
    }
    void operator-=(const var &value) {
        *this = *this - value;
    }
    void operator*=(const var &value) {
        *this = *this * value;
    }
    void operator/=(const var &value) {
        *this = *this / value;
    }    
    
    // array access ////////////////////////////////////////////////////////////
    var &operator[](const int k) {
        switch(_type) {
            case VAR_LIST:
                if (_list) { return (*_list)[k]; }
            break;
            case VAR_STRING:
                { // HACK: this isn't safe
                static var tmp;
                tmp = _string.substr(k,k+1);
                return tmp;
                }
                break;
        }
        static var fail; fail = Error;
        VAR_THROW("UnsupportedError",fail);
    }
    
    var &operator[](const std::string k) {
        switch(_type) {
            case VAR_DICT:
                _dictcheck(1);
                return (*_dict)[k];
            break;
        }
        static var fail; fail = Error;
        VAR_THROW("UnsupportedError",fail);
    }
    
    var &operator[](const char *k) {
        switch(_type) {
            case VAR_DICT:
                _dictcheck(1);
                return (*_dict)[std::string(k)];
            break;
        }
        static var fail; fail = Error;
        VAR_THROW("UnsupportedError",fail);
    }
    
    var &operator[](const var &k) {
        switch(_type) {
            case VAR_LIST:
                if (k._type == VAR_NUMBER) {
                    return (*_list)[(int)k._number];
                }
                break;
            case VAR_DICT:
//                 printf("operator[]\n");
                if (k._type == VAR_STRING) {
                    _dictcheck(1);
                    return (*_dict)[k._string];
                }
//                 return _dict[k.__str__()];
                break;
        }
        static var fail; fail = Error;
        VAR_THROW("UnsupportedError",fail);
    }
    

        
        
    
    
    // VAR_* methods ///////////////////////////////////////////////////////////
    
    const char *c_str() { // not const because _string changes
		// NOTE: must set value to string so that a temporary
		// std::string is not destroyed during the return
		if (_type != VAR_STRING) { _string = __str__(); }
		return _string.c_str();
    }
    
    int __len__() const {
        switch(_type) {
            case VAR_ERROR: return 0; break;
            case VAR_NONE: return 0; break;
            case VAR_NUMBER: return 1; break;
            case VAR_STRING: return _string.size(); break;
            case VAR_LIST: return _list?(*_list).size():0; break;
            case VAR_DICT: return _dict?(*_dict).size():0; break;
        }
        return 0;
    }
    std::string __str__() const {
        switch(_type) {
            case VAR_ERROR:
                return std::string("Error");
                break;
            case VAR_NONE:
                return std::string("None");
                break;
            case VAR_NUMBER: {
                    char s[64];
                    if (_number == ((var_number)(long)_number)) {
                        sprintf(s,"%ld",(long)_number);
                    } else {
                        sprintf(s,"%f",_number);
                        int i,l=strlen(s);
                        for(i=l-1;i>1;i--) {
                            if (s[i] == '0' && s[i-1] != '.') { s[i] = 0; continue; }
                            break;
                        }
                    }
                    return std::string(s);
                    }
                break;
            case VAR_STRING: return _string; break;
            case VAR_LIST: {
                    char s[64];
                    sprintf(s,"<list instance at %p>",this);
                    return std::string(s);
                    }
                break;
            case VAR_DICT: {
                    char s[64];
                    sprintf(s,"<dict instance at %p>",this);
                    return std::string(s);
                    }
                break;
        }
        return std::string("?");
    }
    
    var slice(int a, int b) {
        int l = this->__len__();
        a = (int)var_max(0,(a<0?l+a:a));
        b = (int)var_min(l,(b<0?l+b:b));
        switch(_type) {
            case VAR_STRING:
                return _string.substr(a,b-a);
                break;
            case VAR_LIST: {
                var tmp = var(VAR_LIST,0);
                if (_list) {
                    tmp._listcheck(1);
                    *tmp._list = var_vector<var>(_list->begin()+a,_list->begin()+b);
                }
                return tmp;
                }
                break;
        }
            
        VAR_THROW("UnsupportedError",Error);
    }
    
    bool _dict_has_string(const std::string &v) const {
        return (_type == VAR_DICT && _dict != 0 && _dict->find(v) != _dict->end());
    }
    
    int _list_find_var(const var &v) const {
        VAR_ASSERT(_type == VAR_LIST,"TypeError: method expected a list",-1);
        if (_list == 0) { return -1; }
        var_vector<var>::const_iterator it;
        int n = 0;
        for (it=_list->begin(); it<_list->end(); it++) {
            if (*it == v) { return n; }
            n++;
        }
        return -1;
    }
    
    
    int index(const var &v) const {
        switch(_type) {
            case VAR_LIST: {
                int r = _list_find_var(v);
                VAR_ASSERT(r != -1,"ValueError: list.index(x): x not in list",-1);
                return r;
                }
                break;
            case VAR_STRING: {
                int r = find(v);
                VAR_ASSERT(r != -1,"ValueError: substring not found",-1);
                return r;
                }
                break;
        }
        VAR_THROW("UnsupportedError",-1);
    }
    
    bool has(const var &v) {
        switch(_type) {
            case VAR_DICT:
                VAR_ASSERT(v._type == VAR_STRING,"TypeError: dict keys must be strings",false);
                return _dict_has_string(v._string);
                break;
            case VAR_LIST:
                return _list_find_var(v) != -1;
                break;
            case VAR_STRING:
                return find(v) != -1;
                break;
        }
        VAR_THROW("UnsupportedError",false);
    }
    
    void pop(int v) {
        if (_type == VAR_LIST) {
            int k = v;
            int l = _list?_list->size():0;
            if (k<0) { k = l+k; }
            if (k<0 || k >= l) {
                VAR_THROW("IndexError: list assignment index out of range",);
            }
            _list->erase(_list->begin()+k);
            return;
        }
        VAR_THROW("UnsupportedError",);
    }
    
    void del(const var &v) {
        switch(_type) {
            case VAR_DICT:
                VAR_ASSERT(v._type == VAR_STRING,"TypeError: dict keys must be strings",);
                if (_dict_has_string(v._string)) {
                    _dict->erase(v._string);
                    return;
                }
                VAR_THROW("KeyError",);
                return;
                break;
            case VAR_LIST:
                VAR_ASSERT(v._type == VAR_NUMBER,"TypeError: list indices must be integers",);
                this->pop((int)v._number);
                return;
                break;
        }
        VAR_THROW("UnsupportedError",);
    }
    
    // VAR_DICT methods ////////////////////////////////////////////////////////
    
    var get(const var &v, const var &d) const {
        VAR_ASSERT(_type == VAR_DICT,"UnsupportedError",Error);
    
        VAR_THROW("UnimplementedError",Error);
    }
    
    // VAR_LIST methods ////////////////////////////////////////////////////////
    void append(const var &v) {
        VAR_ASSERT(_type == VAR_LIST,"UnsupportedError",);
        
        _listcheck(1);
        (*_list).push_back(v);
    };
    
    void insert(int k, const var &v) {
        VAR_ASSERT(_type == VAR_LIST,"UnsupportedError",);
            
        _listcheck(1);
        int l = _list?_list->size():0;
        if (k<0) { k = l+k; }
        k = (int)var_max(0,var_min(l-1,k));
        _list->insert(_list->begin()+k,v);
        return;
    }
    
    void extend(const var &v) {
        VAR_ASSERT(_type == VAR_LIST,"UnsupportedError",);
        VAR_ASSERT(v._type == VAR_LIST,"TypeError: lists may only be extended by lists",);
            
        _listcheck(1);
        if (v._list) {
            _list->insert(_list->end(),v._list->begin(),v._list->end());
        }
        return;
    }
    
    void sort() {
        VAR_ASSERT(_type == VAR_LIST,"UnsupportedError",);
            
        if (!_list) { return; }
        std::sort(_list->begin(),_list->end());
        return;
    }
    
    void reverse() {
        VAR_ASSERT(_type == VAR_LIST,"UnsupportedError",);
        
        if (!_list) { return; }
        std::reverse(_list->begin(),_list->end());
        return;
    }
    
    // VAR_STRING methods //////////////////////////////////////////////////////
    
    int find(const var &v) const {
        VAR_ASSERT(_type == VAR_STRING,"UnsupportedError",-1);
        VAR_ASSERT(v._type == VAR_STRING,"TypeError: expected a string",-1);
        
        size_t r = _string.find(v._string);
        if (r == std::string::npos) {
            return -1;
        }
        return r;
    }
    
    var join(const var &v) const {
        VAR_ASSERT(_type == VAR_STRING,"UnsupportedError",Error);
        VAR_ASSERT(v._type == VAR_LIST,"TypeError",Error);
        
        std::string r;
        for (int i=0; i<v.__len__(); i++) {
            if (i != 0) { r += _string; }
            r += (*v._list)[i].__str__();
        }
        return r;
    }
    
    std::string replace(const std::string &k, const std::string &v) const {
        VAR_ASSERT(_type == VAR_STRING,"UnsupportedError","");
        
        std::string s = _string;
        size_t n =0 ; size_t p=0;
        while((n = s.find(k,p)) != std::string::npos) {
            s.replace(n,k.length(),v);
            p = n + v.length();
        }
        return s;
    }
    
    var split(const std::string &delim) const {
        VAR_ASSERT(_type == VAR_STRING,"UnsupportedError","");
        
        var r = var(VAR_LIST,0); r._listcheck(1);
        const std::string &s = _string;
        size_t n =0 ; size_t p=0;
        while((n = s.find(delim,p)) != std::string::npos) {
            r._list->push_back(s.substr(p,n-p));
            p = n + delim.length();
        }
        r._list->push_back(s.substr(p,s.length()-p));
        
        return r;
    }
    
    std::string strip() const {
        VAR_ASSERT(_type == VAR_STRING,"UnsupportedError","");
        
        char const *v = _string.c_str(); int l = _string.length();
        int i; int a = l, b = 0;
        for (i=0; i<l; i++) {
            if (v[i] != ' ' && v[i] != '\n' && v[i] != '\t' && v[i] != '\r') {
                a = (int)var_min(a,i); b = (int)var_max(b,i+1);
            }
        }
        a = (int)var_max(0,(a<0?l+a:a));
        b = (int)var_min(l,(b<0?l+b:b));
        return _string.substr(a,b-a);
    }
};


inline void var_print(const var &v) {
    std::cout << v.__str__();
    std::cout << "\n";
}

inline std::string var_str(const var &v) {
    return v.__str__();
}
inline int var_len(var &v) {
    return v.__len__();
}

inline var var_list() { return var(VAR_LIST,0); }
inline var var_list(var v1) { return var(VAR_LIST,1,&v1); }
inline var var_list(var v1,var v2) { return var(VAR_LIST,2,&v1,&v2); }
inline var var_list(var v1,var v2,var v3) { return var(VAR_LIST,3,&v1,&v2,&v3); }
inline var var_list(var v1,var v2,var v3,var v4) { return var(VAR_LIST,4,&v1,&v2,&v3,&v4); }
inline var var_list(var v1,var v2,var v3,var v4,var v5) { return var(VAR_LIST,5,&v1,&v2,&v3,&v4,&v5); }
inline var var_list(var v1,var v2,var v3,var v4,var v5, var v6) { return var(VAR_LIST,6,&v1,&v2,&v3,&v4,&v5,&v6); }
inline var var_list(var v1,var v2,var v3,var v4,var v5, var v6, var v7) { return var(VAR_LIST,7,&v1,&v2,&v3,&v4,&v5,&v6,&v7); }
inline var var_list(var v1,var v2,var v3,var v4,var v5, var v6, var v7, var v8) { return var(VAR_LIST,8,&v1,&v2,&v3,&v4,&v5,&v6,&v7,&v8); }

inline var var_dict() { return var(VAR_DICT,0); }
inline var var_dict(var k1,var v1) { return var(VAR_DICT,2,&k1,&v1); }
inline var var_dict(var k1,var v1,var k2,var v2) { return var(VAR_DICT,4,&k1,&v1,&k2,&v2); }
inline var var_dict(var k1,var v1,var k2,var v2,var k3,var v3) { return var(VAR_DICT,6,&k1,&v1,&k2,&v2,&k3,&v3); }
inline var var_dict(var k1,var v1,var k2,var v2,var k3,var v3,var k4,var v4) { return var(VAR_DICT,8,&k1,&v1,&k2,&v2,&k3,&v3,&k4,&v4); }
inline var var_dict(var k1,var v1,var k2,var v2,var k3,var v3,var k4,var v4,var k5,var v5) { return var(VAR_DICT,10,&k1,&v1,&k2,&v2,&k3,&v3,&k4,&v4,&k5,&v5); }
inline var var_dict(var k1,var v1,var k2,var v2,var k3,var v3,var k4,var v4,var k5,var v5,var k6,var v6) { return var(VAR_DICT,12,&k1,&v1,&k2,&v2,&k3,&v3,&k4,&v4,&k5,&v5,&k6,&v6); }
inline var var_dict(var k1,var v1,var k2,var v2,var k3,var v3,var k4,var v4,var k5,var v5,var k6,var v6,var k7,var v7) { return var(VAR_DICT,14,&k1,&v1,&k2,&v2,&k3,&v3,&k4,&v4,&k5,&v5,&k6,&v6,&k7,&v7); }
inline var var_dict(var k1,var v1,var k2,var v2,var k3,var v3,var k4,var v4,var k5,var v5,var k6,var v6,var k7,var v7,var k8,var v8) { return var(VAR_DICT,16,&k1,&v1,&k2,&v2,&k3,&v3,&k4,&v4,&k5,&v5,&k6,&v6,&k7,&v7,&k8,&v8); }

// struct varArgs {
//     std::vector<var> args;
//     varArgs() { }
// 
//     varArgs& operator,(const var &v) {
//         args.push_back(v);
//         return *this;
//     }
//     
//     operator var() {
//         var r = var(VAR_LIST,0);
//         r._listcheck(1);
//         *r._list = args;
//         return r;
//     }
//         
// };



struct var_iter {
    int n;
    int type;
    var_vector<var>::iterator list_it,list_end;
    std::map<std::string,var>::iterator dict_it,dict_end;
    std::string::iterator string_it,string_end;
    var &value;
    var tmp;
    
    // lots of var copying, not very efficient
    var_iter(var &_value) : value(_value) {
        n=0;
        type = value._type;
        switch(type) {
            case VAR_LIST:
                list_it = value._list->begin();
                list_end = value._list->end();
                break;
            case VAR_DICT:
                dict_it = value._dict->begin();
                dict_end = value._dict->end();
                break;
            case VAR_STRING:
                string_it = value._string.begin();
                string_end = value._string.end();
                break;
            default:
                VAR_THROW("TypeError: object is not iterable",);
                break;
        }
    }
    
    bool first() {
        return n == 0;
    }
    
    bool next(var **v) {
        bool r = false;
        switch(type) {
            case VAR_LIST:
                if (list_it != list_end) {
                    *v = &(*list_it);
                    list_it ++;
                    r = true;
                }
                break;
            case VAR_DICT:
                if (dict_it != dict_end) {
                    tmp = (*dict_it).first;
                    *v = &tmp;
                    dict_it ++;
                    r = true;
                }
                break;
            case VAR_STRING:
                if (string_it != string_end) {
                    tmp = (*string_it);
                    *v = &tmp;
                    string_it ++;
                    r = true;
                }
                break;
        }
        n += 1;
        return r;
    }
    bool pair(var &k, var **v) {
        bool r = false;
        switch(type) {
            case VAR_LIST:
                if (list_it != list_end) {
                    k = n;
                    *v = &(*list_it);
                    list_it ++;
                    r = true;
                }
                break;
            case VAR_DICT:
                if (dict_it != dict_end) {
                    k = (*dict_it).first;
                    *v = &((*dict_it).second);
                    dict_it ++;
                    r = true;
                }
                break;
            case VAR_STRING:
                if (string_it != string_end) {
                    k = n;
                    tmp = (*string_it);
                    *v = &tmp;
                    string_it ++;
                    r = true;
                }
                break;
        }
        n += 1;
        return r;
    }
};

#define var_for_in(v,args) \
    for(var_iter _iter(args); _iter.first(); ) \
    for(var *_v; _iter.next(&_v); ) \
    for(var &v=*_v; _v!=0; _v=0)

#define var_pair_in(k,v,args) \
    for(var_iter _iter(args); _iter.first(); ) \
    for(var k,*_v; _iter.pair(k,&_v); ) \
    for(var &v=*_v; _v!=0; _v=0)



#ifdef VAR_TEST

int main(int argc, char *argv[]) {

    // test None, number, string
    assert(var_str(None) == "None");
    assert(var_str(3) == "3");
    assert(var_str("test") == "test");
    assert(var_str(3.5) == "3.5");
    
    // test list
    var x = list(None,3,"test",3.5);
    assert(len(x) == 4);
    assert(var_str(x[0]) == "None");
    assert(var_str(x[1]) == "3");
    assert(var_str(x[2]) == "test");
    assert(var_str(x[3]) == "3.5");
    x = list();
    
    // test dict
    var y = dict("a",None,"b",3,"c","test","d",3.5);
    assert(len(y) == 4);
    assert(var_str(y["a"]) == "None");
    assert(var_str(y["b"]) == "3");
    assert(var_str(y["c"]) == "test");
    assert(var_str(y["d"]) == "3.5");
    y = dict();

	// test nested dict
	y = dict("key1",dict("key2","value"));
	assert(strcmp(y["dict1"]["dict2"].c_str(),"value")==0);
	y = dict();
    
    // test list+dict
    var z = list(dict("a",list("X","OK")));
    assert(var_str(z[0]["a"][1]) == "OK");
    
    // test number addition
    var a,b,c;
    a = 2; b = 3; c = a + b;
    assert(c == 5);
    
    // test string addition
    a = "O"; b = "K"; c = a + b;
    assert(c == "OK");
    
    // test list addition
    a = list(2); b = list(3);
    c = a + b;
    assert(len(c) == 2);
    assert(c[0] == 2);
    assert(c[1] == 3);

    // test string slice
    a = "xOKx";
    b = a.slice(1,3);
    assert(b == "OK");
    
    // test list slice
    a = list(0,"O","K",1);
    b = a.slice(1,3);
    assert(b[0] == "O");
    assert(b[1] == "K");
    
    // test dict has
    a = dict("k","v");
    assert(a.has("k") == true);
    assert(a.has("v") == false);
    
    // test dict del
    a = dict("k","v");
    assert(a.has("k") == true);
    a.del("k");
    assert(a.has("k") == false);
    
    // test list del / pop
    a = list(1,2,3);
    assert(len(a) == 3);
    a.del(1);
    assert(len(a) == 2);
    assert(a[0] == 1);
    assert(a[1] == 3);
    
    // test list has
    a = list(1,2,3);
    assert(a.has(1) == true);
    assert(a.has(17) == false);
    
    // test list index
    a = list(2,3,4);
    assert(a.index(2) == 0);
    assert(a.index(3) == 1);
    assert(a.index(4) == 2);
    assert(a.index(1) == -1);
    
    // test list insert
    a = list(1,2,3);
    a.insert(1,5);
    assert(len(a) == 4);
    assert(a[0] == 1);
    assert(a[1] == 5);
    assert(a[2] == 2);
    
    // test list extend
    a = list(1,2);
    assert(len(a)==2);
    a.extend(list(3,4));
    assert(len(a)==4);
    assert(a[0]==1);
    assert(a[1]==2);
    assert(a[2]==3);
    assert(a[3]==4);
    
    // test list sort
    a = list(3,1,2);
    a.sort();
    assert(a[0]==1);
    assert(a[1]==2);
    assert(a[2]==3);
    
    // test string c_str
    a = "test";
    assert(strcmp(a.c_str(),"test")==0);
    
    // test string join
    a = list("O","K");
    b = var("-").join(a);
    assert(b == "O-K");
    
    // test string index / find / has
    a = "-OK-";
    assert(a.index("OK") == 1);
    assert(a.index("X") == -1);
    check_error("ValueError");
    assert(a.find("OK") == 1);
    assert(a.find("X") == -1);
    assert(a.has("OK") == true);
    assert(a.has("X") == false);
    
    // test list reverse
    a = list(1,2,3);
    a.reverse();
    assert(a[0] == 3);
    assert(a[1] == 2);
    assert(a[2] == 1);
    
    // test string replace
    a = "-O-K-";
    b = a.replace("-","X");
    assert(b == "XOXKX");
    
    // test string split
    a = "O-K";
    b = a.split("-");
    assert(b[0] == "O");
    assert(b[1] == "K");
    
    // test string strip
    a = " OK ";
    b = a.strip();
    assert(b == "OK");
    
    // test string[]
    a = "OK";
    assert(a[0] == "O");
    assert(a[1] == "K"); 
    
    // test string !=
    a = "O";
    b = "K";
    assert(a != b);
    
    // test list iter
    a = list(1,2,3);
    b = 0;
    for_in(c,a) {
        b = b + c;
    }
    assert(b == 6);
    
    // test dict iter for_in
    a = dict("a",1,"b",2,"c",3);
    b = 0;
    for_in(e,a) {
        b = b + a[e];
    }
    assert(b == 6);
    
    // test dict iter pair_in
    a = dict("a",1,"b",2,"c",3);
    b = "";
    pair_in(k,v,a) {
        b = b + k + var_str(v);
    }
    assert(b == "a1b2c3");
    
    // test inline operators
    a = 3;
    a += 4;
    assert(a == 7);
    a *= 3;
    assert(a == 21);
    a -= 1;
    assert(a == 20);
    a /= 2;
    assert(a == 10);
    
    // test conversions back to std::string, float, int, double
    a = 5.2;
    float _float = a;
    assert(var_abs(_float-5.2) < 0.001);
    int _int = a;
    assert(_int == 5);
    double _double = a;
    assert(var_abs(_double-5.2) < 0.001);
    b = "test";
    std::string _string = b;
    assert(b == "test");
    
    
    // test all error conditions
    // document it
    // list:sort_by
    
    printf("sizeof(var) %d, var_number %d, std::string %d, list: %d, dict: %d\n",
        sizeof(var),
        sizeof(var_number),
        sizeof(std::string),
        sizeof(var_vector<var>),
        sizeof(var_map<std::string,var>));

    return 0;

}
#endif

#endif
