#ifdef GC_BUILD_C
extern "C"
{
#endif

/*
** client.c -- a stream socket client demo
*/

// http://www.jmarshall.com/easy/http/

#include <stdio.h>
#include <stdlib.h>
#ifndef PLATFORM_WIN32
#include <unistd.h>
#endif
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>

#include "cuz_web.h"

#define GC_WEB_MAXPACKET 4096

#ifndef PLATFORM_WIN32
#include <sys/time.h>
#endif
// #include <unistd.h>

void cuz_web_extend(cuz_web_buffer *buf, int ext) {
    int target = buf->len + ext;
    int alloc = buf->alloc?buf->alloc:1;
    while (alloc < target) { alloc *= 2; }
    if (buf->alloc >= alloc) { return; }
//     fprintf(stderr,"cuz_web_extend: %d .. %d->%d\n",buf->len,buf->alloc,alloc);
    buf->alloc = alloc;
    buf->data = (unsigned char*)realloc(buf->data,buf->alloc);
    memset(buf->data+buf->len,0,buf->alloc-buf->len); // reset it to 0
}

// #ifdef BUILD_SDL
// long cuz_web_get_ticks() {
//     return SDL_GetTicks();
// }
// void cuz_web_delay(long ticks) {
//     SDL_Delay(ticks);
// }
// #else
// 
// // HACK: this function is not thread-safe
// long cuz_web_get_ticks() {
//     static int cuz_timer_is_init = 0;
//     static struct timeval cuz_timer_s;
//     if (!cuz_timer_is_init) {
//         cuz_timer_is_init = 1;
//         gettimeofday(&cuz_timer_s, NULL);
//     }
//     struct timeval t;
//     gettimeofday(&t, NULL);
//     long r = (t.tv_sec-cuz_timer_s.tv_sec)*1000 + (t.tv_usec-cuz_timer_s.tv_usec) / 1000;
//     return r;
// }
// 
// void cuz_web_delay(long ticks) {
//     struct timespec a, b;
//     a.tv_sec = ticks / 1000;
//     a.tv_nsec = (ticks % 1000) * 1000000;
//     b.tv_sec = a.tv_sec;
//     b.tv_nsec = a.tv_nsec; 
//     nanosleep(&a,&b);
// }
// #endif

long cuz_web_get_ticks() {
    return time(0);
}

void cuz_web_delay(long ticks) { }

int cuz_web_urlencode_len(const char *in, int l) {
    int total = 0;
    int i;
    for (i=0; i<l; i++) {
        char c = in[i];
        if ((c >= 'a' && c <= 'z') ||
            (c >= 'A' && c <= 'Z') ||
            (c >= '0' && c <= '9') ) {
            total += 1;
            continue;
        }
        total += 3;
    }
    return total+1;
}

void cuz_web_urlencode(char *out, int len, const char *in, int l) {
    int i;
    int n =0; unsigned char c;
    const char *hex = "0123456789abcdef";
    memset(out,0,len);
    for (i=0; i<l; i++) {
        c = in[i];
        if ((c >= 'a' && c <= 'z') ||
            (c >= 'A' && c <= 'Z') ||
            (c >= '0' && c <= '9') ) {
            if (n > (len-2)) { break; }
            out[n++] = c;
            continue;
        }
        if (n > (len-4)) { break; }
        out[n++] = '%';
        out[n++] = hex[c>>4];
        out[n++] = hex[c&0xf];
    }
}


void cuz_web_reset_timeout(cuz_web *page) {
    page->timeout = cuz_web_get_ticks() + page->_timeout;
}

    // init and deinit taken from enet
int _cuz_web_is_init = 0;
int cuz_web_init() {
#ifdef PLATFORM_WIN32
    WORD versionRequested = MAKEWORD (1, 1);
    WSADATA wsaData;
   
	if (WSAStartup (versionRequested, & wsaData)) {
       return -1;
	}

    if (LOBYTE (wsaData.wVersion) != 1||
        HIBYTE (wsaData.wVersion) != 1)
    {
       WSACleanup ();
       
       return -1;
    }
#endif
	_cuz_web_is_init = 1;
    return 0;
}

void cuz_web_deinit() {
	if (!_cuz_web_is_init) { return; }
#ifdef PLATFORM_WIN32
    WSACleanup ();
#endif
}


cuz_web *cuz_web_create(const char *_host, int port, const char *_path, const void *post, int postlen, int timeout) {
    cuz_web *page = (cuz_web*)calloc(sizeof(cuz_web),1);
    
    page->state = GC_WEB_INIT;
    page->data = 0; page->len = 0; page->error = 0;
    strcpy(page->host,_host); page->port = port; strcpy(page->path,_path);
    
//     fprintf(stderr,"%s\n",path);
    
//     page->_timeout = clock() + (timeout * CLOCKS_PER_SEC);
    page->_timeout = timeout; // *1000, if we're doing SDL timers ..
    cuz_web_reset_timeout(page);
    
    page->sockfd = -1; // keeps valgrind quiet, hopefully won't cause some bizarre bug that will destroy my life
    
    char request[1024];
    cuz_web_extend(&page->_request,1024);
    if (!postlen) {
        sprintf((char*)page->_request.data,"GET %s HTTP/1.0\nHost: %s:%d\n\n",page->path,page->host,page->port);
        page->_request.len = strlen((char*)page->_request.data);
    } else {
        sprintf((char*)page->_request.data,"POST %s HTTP/1.0\nHost: %s:%d\nContent-Type: application/x-www-form-urlencoded\nContent-Length: %d\n\n",page->path,page->host,page->port,postlen);
        
        page->_request.len = strlen((char*)page->_request.data);
        cuz_web_extend(&page->_request,postlen);
        memcpy(page->_request.data+page->_request.len,post,postlen);
        page->_request.len += postlen;
//         fprintf(stderr,"request: %s\n",page->_request.data);
    }
    

    // NOTE: we should switch to getaddrinfo, but i'm not sure how
    struct hostent *he;
    if ((he=gethostbyname(page->host)) == NULL) {  // get the host info 
        #ifndef PLATFORM_WIN32
            herror("gethostbyname");
        #endif
		fprintf(stderr,"gethostbyname(%s) == null; h_errno = %d\n",page->host,h_errno);
        page->error = 1;
        return page;
    }
    // so since this function is non-reentrant, and it's return value
    // needs to be used immediately ...
    page->their_addr.sin_family = AF_INET;    // host byte order 
    page->their_addr.sin_port = htons(page->port);  // short, network byte order 
    page->their_addr.sin_addr = *((struct in_addr *)he->h_addr);
    memset(page->their_addr.sin_zero, '\0', sizeof page->their_addr.sin_zero);

    if ((page->sockfd = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
        perror("socket");
		fprintf(stderr,"socket() == -1\n");
        page->error = 1;
        return page;
    }
    
    #ifndef PLATFORM_WIN32
        fcntl(page->sockfd, F_SETFL, O_NONBLOCK);
    #else
        /*
        u_long nonBlocking = 1;
        int result = ioctlsocket (page->sockfd, FIONBIO, & nonBlocking);
        if (result) {
            int wsa_error = WSAGetLastError();
            fprintf(stderr,"ioctlsocket::wsaerror::%d\n",
                wsa_error);
            page->error = 1;
            return page;
        }
        */
    #endif
    
    page->state = GC_WEB_CONNECT;

    return page;
}

void cuz_find_header(cuz_web *page) {
    int i; char p=0,c;
    page->header = page->_response.len;
    for (i=0; i<page->_response.len; i++) {
        c = ((char*)page->_response.data)[i];
        if (c == '\r') { continue; } // most web responses include these, ignore them
        if (p == '\n' && c == '\n') {
            page->header = i+1;
            return;
        }
        p = c;
    }
}
    
// doesn't handle multi-line header data properly
void cuz_web_get_header(cuz_web *page, char *dest, int dlen, const char *name) {
    int n = 0;
    char s[256];
    char *d = (char*)page->_response.data;
    int l = page->header;
    int i;
    int len;
    sprintf(s,"%s:",name);
    len = strlen(s);
    memset(dest,0,dlen);

    while (1) {
        if ((n+len) >= l) { break; }
        if (strncmp(d+n,s,len)==0) {
            n += len;
            while (n<l && d[n] == ' ') { n += 1; }
            i = 0;
            while (n<l && i < (dlen-1) && d[n] != '\r' && d[n] != '\n') {
                dest[i] = d[n];
                i += 1; n += 1;
            }
            return;
        }
        while (n<l && d[n] != '\n') {
            //fprintf(stderr,"%d < %d\n",n,l);
            n += 1;
        }
        if (n<l) { n += 1; }
    }
}

int cuz_web_poll(cuz_web *page) {
    if (page->state == GC_WEB_CLOSED) { return 1; }
    
    
    if (page->error) {
//         fprintf(stderr,"error condition already met\n");
        return 1;
    }
    
    if (cuz_web_get_ticks() > page->timeout) {
        fprintf(stderr,"timeout :: state = %d\n",page->state);
        page->error = 1;
        return 1;
    }
    
    if (page->state == GC_WEB_CONNECT) {
            if (connect(page->sockfd, (struct sockaddr *)&page->their_addr,
                                                sizeof page->their_addr) != 0) {
                if (errno != EINPROGRESS) {
                    fprintf(stderr,"connect: %d\n",errno);
                    perror("connect");
                    page->error = 1;
                    return 1;
                }
            }
            page->state = GC_WEB_SEND;
            cuz_web_reset_timeout(page);
    }

    
    fd_set readfds, writefds, errorfds;
    struct timeval tv;

    // clear the set ahead of time
    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    FD_ZERO(&errorfds);

    // add our descriptors to the set
    FD_SET(page->sockfd, &readfds);
    FD_SET(page->sockfd, &writefds);
    FD_SET(page->sockfd, &errorfds);
    

    // wait until either socket has data ready to be recv()d 
    tv.tv_sec = 0; tv.tv_usec = 0;
    int rv = select(page->sockfd+1, &readfds, &writefds, &errorfds, &tv);

    if (rv == -1) {
        perror("select"); // error occurred in select()
        fprintf(stderr,"error in select\n");
        page->error = 1;
        return 1;
    } else if (rv == 0) {
        // fprintf(stderr,"S");
        return 0;
    }
    
    //fprintf(stderr,"state: %d\n",page->state);
     
    switch (page->state) {
        case GC_WEB_CONNECT: {
        }
        break;

        case GC_WEB_SEND: {
            //fprintf(stderr,"about to send: %s\n",page->_request.data);
            if (!FD_ISSET(page->sockfd, &writefds)) { return 0; }

            int numbytes;
            if ((numbytes=send(page->sockfd, (char*)page->_request.data+page->_request.pos, page->_request.len-page->_request.pos, 0)) == -1) {
                perror("send");
                fprintf(stderr,"send error!!\n");
                page->error = 1;
                return 1;
            }
            page->_request.pos += numbytes;
            //fprintf(stderr,"sent: %d\n",numbytes);
            if (numbytes) { cuz_web_reset_timeout(page); }
            if (page->_request.pos >= page->_request.len) {
                page->state = GC_WEB_RECEIVE;
            }
        }
        break;
        
        case GC_WEB_RECEIVE: {
            //fprintf(stderr,"we are about to receive\n");
            if (!FD_ISSET(page->sockfd, &readfds)) { return 0; }
            
            cuz_web_extend(&page->_response,GC_WEB_MAXPACKET);
            page->data = page->_response.data + page->header;
            
            int numbytes;
            if ((numbytes=recv(page->sockfd, (char*)page->_response.data+page->_response.len, GC_WEB_MAXPACKET, 0)) == -1) {
                perror("recv");
                page->error = 1;
                return 1;
            }
            
            // test if the connection is closed
            if (numbytes == 0) {
                 //fprintf(stderr,"%s",page->data);
                 //fprintf(stderr,"\nclosed!!\n");
                page->state = GC_WEB_CLOSED;
                return 1;
            }
            
            page->_response.len += numbytes;
            if (numbytes) { cuz_web_reset_timeout(page); }
            if (!page->header) { cuz_find_header(page); }
            page->data = page->_response.data + page->header;
            page->len = page->_response.len - page->header;

        }
        break;
        
        case GC_WEB_CLOSED:
            return 1;
        break;
    }
    
    
    return 0;

}
void cuz_web_destroy(cuz_web *page) {
    if (page->sockfd != -1) {
#ifndef PLATFORM_WIN32
		close(page->sockfd);
#else
		closesocket(page->sockfd);
#endif
	}
	free(page->_request.data);
	free(page->_response.data);
    free(page);
}



#ifdef GC_BUILD_C
}
#endif


#ifdef BUILD_TEST_WEB

#define TEST_TIMEOUT 5.0

#define ASSERT(v) \
    if (!(v)) { \
        fprintf(stderr,"ASSERT FAILED [%s:%d] %s(): %s\n",__FILE__,__LINE__,__func__,#v); \
        exit(-1); \
    }
    
#define TEST_INFO \
    fprintf(stderr,"\n\n===================================================\n"); \
    fprintf(stderr,"[%s:%d] %s() ...\n",__FILE__,__LINE__,__func__); 



void test_web() {
    TEST_INFO;
    cuz_web *page;
    page = cuz_web_create("www.google.com",80,"/",0,0,TEST_TIMEOUT);
    while (!cuz_web_poll(page)) { cuz_web_delay(1); }
//      fprintf(stderr,"%s\n",page->data);
    ASSERT(page->error == 0);
    ASSERT(strstr(page->data,"Google") != 0);
    ASSERT(strstr(page->data,"</script>") != 0);
    cuz_web_destroy(page);
}

void test_web_skip_headers() {
    TEST_INFO;
    cuz_web *page;
    page = cuz_web_create("www.google.com",80,"/",0,0,TEST_TIMEOUT);
    while (!cuz_web_poll(page)) { cuz_web_delay(1); }
//      fprintf(stderr,"%s\n",page->data);
    ASSERT(page->error == 0);
    ASSERT(strstr(page->data,"HTTP/1.") == 0); // we want to skip the headers
    cuz_web_destroy(page);
}


void test_web_post() {
    TEST_INFO;
    cuz_web *page;
    const char *post = "q=test";
    page = cuz_web_create("www.example.com",80,"/webtest",post,strlen(post),TEST_TIMEOUT);
    while (!cuz_web_poll(page)) { cuz_web_delay(1); }
    ASSERT(page->error == 0);
    ASSERT(strstr(page->data,"Bad Request") == 0);
    ASSERT(strstr(page->data,"404 Not Found") != 0);
    cuz_web_destroy(page);
}

void test_web_bad_dns() {
    TEST_INFO;
    cuz_web *page;
    page = cuz_web_create("asdlkmasdcm34krbtla54sdcasdfasdaa.com",80,"/",0,0,TEST_TIMEOUT);
    while (!cuz_web_poll(page)) { cuz_web_delay(1); }
    ASSERT(page->error != 0);
    cuz_web_destroy(page);
}

void test_web_bad_ip() {
    TEST_INFO;
    cuz_web *page;
    page = cuz_web_create("192.168.2.99",80,"/",0,0,TEST_TIMEOUT);
    while (!cuz_web_poll(page)) { cuz_web_delay(1); }
    ASSERT(page->error != 0);
    cuz_web_destroy(page);
}

void test_web_bad_port() {
    TEST_INFO;
    cuz_web *page;
    page = cuz_web_create("www.galcon.com",9876,"/",0,0,TEST_TIMEOUT);
    while (!cuz_web_poll(page)) { cuz_web_delay(1); }
    ASSERT(page->error != 0);
    cuz_web_destroy(page);
}

void test_web_huge_download() {
    TEST_INFO;
    cuz_web *page;
    page = cuz_web_create("www.imitationpickles.org",80,"/galcon/files/Galcon.dmg",0,0,TEST_TIMEOUT);
    while (!cuz_web_poll(page)) { cuz_web_delay(1); }
    ASSERT(page->error != 0);
    ASSERT(page->len != 0);
    cuz_web_destroy(page);
}

void test_web_multi_download() {
    TEST_INFO;
    cuz_web *p1; cuz_web *p2;
    p1 = cuz_web_create("www.imitationpickles.org",80,"/galcon/files/Galcon.dmg",0,0,TEST_TIMEOUT);
    p2 = cuz_web_create("www.imitationpickles.org",80,"/galcon/files/galcon.tgz",0,0,TEST_TIMEOUT);
    while ((cuz_web_poll(p1) + cuz_web_poll(p2)) < 2) { cuz_web_delay(1); }
    ASSERT(p1->error != 0); ASSERT(p1->len != 0);
    ASSERT(p2->error != 0); ASSERT(p2->len != 0);
    
    cuz_web_destroy(p1);
    cuz_web_destroy(p2);
}

void test_web_urlencode() {
    TEST_INFO;
    char *in = "a b c";
    char out[256];
    cuz_web_urlencode(out,256,in,strlen(in));
    ASSERT(strcmp(out,"a%20b%20c")==0);
}

void test_web_get_header() {
    TEST_INFO;
    cuz_web *page;
    page = cuz_web_create("www.google.com",80,"/",0,0,TEST_TIMEOUT);
    while (!cuz_web_poll(page)) { cuz_web_delay(1); }
    //fprintf(stderr,"%s\n",page->_response.data);
    ASSERT(page->error == 0);
    char s[256];
    cuz_web_get_header(page,s,256,"Content-Type");
    ASSERT(strcmp(s,"text/html; charset=ISO-8859-1") == 0);
    cuz_web_destroy(page);
}

int main(int argc, char *argv[]) {

    test_web_urlencode();
    test_web();
    test_web_skip_headers();
    test_web_get_header();
    test_web_post();
    test_web_bad_dns();
    test_web_bad_ip();
    test_web_bad_port();
    test_web_huge_download();
    test_web_multi_download();
}

//  gcc -g -D BUILD_TEST_WEB web.c -o webtest

#endif
